/********************************************************************************
*                                                                               *
*                      F X   M O V I E   M A N A G E R                          *
*                                                                               *
*********************************************************************************
* Copyright(C) 2008-2014 by Luigi Ziviani <ziviani_luigi@yahoo.it>.             *
* All Rights Reserved.                                 v 6.6                    *
*********************************************************************************
* Id: fxplayer_filters.c,                              2008/01/01               *
*********************************************************************************
* This file is part of FXMOVIEMANAGER                                           *
*                                                                               *
* FXMOVIEMANAGER is free software; you can redistribute it and/or modify        *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation; either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* FXMOVIEMANAGER is distributed in the hope that it will be useful,             *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with this program; if not, write to the Free Software                   *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA     * 
********************************************************************************/

#include "fxplayer_filters.h"

#ifdef ENABLE_PLAYERS

#if HAVE_LIBAVFILTER
int configure_filtergraph(AVFilterGraph *graph, const char *filtergraph,
                                 AVFilterContext *source_ctx, AVFilterContext *sink_ctx)
{
    int ret;
    AVFilterInOut *outputs = NULL, *inputs = NULL;

    if (filtergraph) {
        outputs = avfilter_inout_alloc();
        inputs  = avfilter_inout_alloc();
        if (!outputs || !inputs) {
            ret = AVERROR(ENOMEM);
            goto fail;
        }

        outputs->name       = av_strdup("in");
        outputs->filter_ctx = source_ctx;
        outputs->pad_idx    = 0;
        outputs->next       = NULL;

        inputs->name        = av_strdup("out");
        inputs->filter_ctx  = sink_ctx;
        inputs->pad_idx     = 0;
        inputs->next        = NULL;

        if ((ret = avfilter_graph_parse_ptr(graph, filtergraph, &inputs, &outputs, NULL)) < 0)
            goto fail;
    } else {
        if ((ret = avfilter_link(source_ctx, 0, sink_ctx, 0)) < 0)
            goto fail;
    }

    ret = avfilter_graph_config(graph, NULL);
fail:
    avfilter_inout_free(&outputs);
    avfilter_inout_free(&inputs);
    return ret;
}


int configure_video_filters(AVFilterGraph *graph, VideoState *vs, const char *vfilters, AVFrame *frame)
{
    static const enum AVPixelFormat pix_fmts[] = { PIX_FMT_YUV420P, PIX_FMT_NONE };
    char sws_flags_str[128];
    char buffersrc_args[256];
    int ret;
 
    AVFilterContext *filt_src = NULL, *filt_out = NULL, *filt_crop;
    AVCodecContext *codec = vs->video_st->codec;
    AVRational fr = av_guess_frame_rate(vs->avf_ctx, vs->video_st, NULL);

    av_opt_get_int(sws_opts, "sws_flags", 0, &sws_flags);
    snprintf(sws_flags_str, sizeof(sws_flags_str), "flags=%"PRId64, sws_flags);
    graph->scale_sws_opts = av_strdup(sws_flags_str);

    snprintf(buffersrc_args, sizeof(buffersrc_args),
             "video_size=%dx%d:pix_fmt=%d:time_base=%d/%d:pixel_aspect=%d/%d",
             frame->width, frame->height, frame->format,
             vs->video_st->time_base.num, vs->video_st->time_base.den,
             codec->sample_aspect_ratio.num, FFMAX(codec->sample_aspect_ratio.den, 1));

    if (fr.num && fr.den)
        av_strlcatf(buffersrc_args, sizeof(buffersrc_args), ":frame_rate=%d/%d", fr.num, fr.den);

    if ((ret = avfilter_graph_create_filter(&filt_src,
                                            avfilter_get_by_name("buffer"),
                                            "fx_buffer", buffersrc_args, NULL,
                                            graph)) < 0){
        av_log(NULL, AV_LOG_INFO, "failed avfilter_graph_create_filter: buffer\n");
        goto fail;
    }

    ret = avfilter_graph_create_filter(&filt_out,
                                       avfilter_get_by_name("buffersink"),
                                       "fx_buffersink", NULL, NULL, graph);
    if (ret < 0){
        av_log(NULL, AV_LOG_INFO, "failed avfilter_graph_create_filter: buffersink\n");
        goto fail;
    }

    if ((ret = av_opt_set_int_list(filt_out, "pix_fmts", pix_fmts,  AV_PIX_FMT_NONE, AV_OPT_SEARCH_CHILDREN)) < 0){
        av_log(NULL, AV_LOG_INFO, "failed av_opt_set_int_list: pix_fmts\n");
        goto fail;
    }


    /* SDL YUV code is not handling odd width/height for some driver
     * combinations, therefore we crop the picture to an even width/height. */
    if ((ret = avfilter_graph_create_filter(&filt_crop,
                                            avfilter_get_by_name("crop"),
                                            "fx_crop", "floor(in_w/2)*2:floor(in_h/2)*2", NULL, graph)) < 0){
        av_log(NULL, AV_LOG_INFO, "failed avfilter_graph_create_filter: crop\n");
        goto fail;
    }

    if ((ret = avfilter_link(filt_crop, 0, filt_out, 0)) < 0){
        av_log(NULL, AV_LOG_INFO, "failed avfilter_link\n");
        goto fail;
    }


    if ((ret = configure_filtergraph(graph, vfilters, filt_src, filt_crop)) < 0){
        av_log(NULL, AV_LOG_INFO, "failed configure_filtergraph: video\n");
        goto fail;
    }

    vs->in_video_filter  = filt_src;
    vs->out_video_filter = filt_out;

fail:
    return ret;
}


int configure_audio_filters(VideoState *vs, const char *afilters, int force_output_format)
{
    static const enum AVSampleFormat sample_fmts[] = { AV_SAMPLE_FMT_S16, AV_SAMPLE_FMT_NONE };
    int sample_rates[2] = { 0, -1 };
    int64_t channel_layouts[2] = { 0, -1 };
    int channels[2] = { 0, -1 };
    AVFilterContext *filt_asrc = NULL, *filt_asink = NULL;
    char asrc_args[256];
    int ret;

    avfilter_graph_free(&vs->agraph);
    if (!(vs->agraph = avfilter_graph_alloc()))
        return AVERROR(ENOMEM);

    ret = snprintf(asrc_args, sizeof(asrc_args),
                   "sample_rate=%d:sample_fmt=%s:channels=%d:time_base=%d/%d",
                   vs->audio_filter_src.freq, av_get_sample_fmt_name(vs->audio_filter_src.fmt),
                   vs->audio_filter_src.channels,
                   1, vs->audio_filter_src.freq);
    if (vs->audio_filter_src.channel_layout)
        snprintf(asrc_args + ret, sizeof(asrc_args) - ret,
                 ":channel_layout=0x%"PRIx64,  vs->audio_filter_src.channel_layout);


    ret = avfilter_graph_create_filter(&filt_asrc,
                                       avfilter_get_by_name("abuffer"), "fx_abuffer",
                                       asrc_args, NULL, vs->agraph);
    if (ret < 0){
        av_log(NULL, AV_LOG_INFO, "failed avfilter_graph_create_filter: abuffer\n");
        goto end;
    }


    ret = avfilter_graph_create_filter(&filt_asink,
                                       avfilter_get_by_name("abuffersink"), "fx_abuffersink",
                                       NULL, NULL, vs->agraph);
    if (ret < 0){
        av_log(NULL, AV_LOG_INFO, "failed avfilter_graph_create_filter: abuffersink\n");
        goto end;
    }


    if ((ret = av_opt_set_int_list(filt_asink, "sample_fmts", sample_fmts,  AV_SAMPLE_FMT_NONE, AV_OPT_SEARCH_CHILDREN)) < 0)
        goto end;
    if ((ret = av_opt_set_int(filt_asink, "all_channel_counts", 1, AV_OPT_SEARCH_CHILDREN)) < 0)
        goto end;

    if (force_output_format) {
        channel_layouts[0] = vs->audio_tgt.channel_layout;
        channels       [0] = vs->audio_tgt.channels;
        sample_rates   [0] = vs->audio_tgt.freq;
        if ((ret = av_opt_set_int(filt_asink, "all_channel_counts", 0, AV_OPT_SEARCH_CHILDREN)) < 0)
            goto end;
        if ((ret = av_opt_set_int_list(filt_asink, "channel_layouts", channel_layouts,  -1, AV_OPT_SEARCH_CHILDREN)) < 0)
            goto end;
        if ((ret = av_opt_set_int_list(filt_asink, "channel_counts" , channels       ,  -1, AV_OPT_SEARCH_CHILDREN)) < 0)
            goto end;
        if ((ret = av_opt_set_int_list(filt_asink, "sample_rates"   , sample_rates   ,  -1, AV_OPT_SEARCH_CHILDREN)) < 0)
            goto end;
    }


    if ((ret = configure_filtergraph(vs->agraph, afilters, filt_asrc, filt_asink)) < 0){
        av_log(NULL, AV_LOG_INFO, "failed configure_filtergraph: audio\n");
        goto end;
    }


    vs->in_audio_filter  = filt_asrc;
    vs->out_audio_filter = filt_asink;

end:
    if (ret < 0)
        avfilter_graph_free(&vs->agraph);
    return ret;
}
#endif /* HAVE_LIBAVFILTER */

#endif /* ENABLE_PLAYERS */

