/********************************************************************************
*                                                                               *
*                      F X   M O V I E   M A N A G E R                          *
*                                                                               *
*********************************************************************************
* Copyright(C) 2008-2014 by Luigi Ziviani <ziviani_luigi@yahoo.it>.             *
* All Rights Reserved.                                 v 6.6                    *
*********************************************************************************
* Id: fxplayer_filters.h,                              2008/01/01               *
*********************************************************************************
* This file is part of FXMOVIEMANAGER                                           *
*                                                                               *
* FXMOVIEMANAGER is free software; you can redistribute it and/or modify        *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation; either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* FXMOVIEMANAGER is distributed in the hope that it will be useful,             *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with this program; if not, write to the Free Software                   *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA     * 
********************************************************************************/


#ifndef FXPLAYER_FILTERS_H
#define FXPLAYER_FILTERS_H

#include "config.h"

#ifdef ENABLE_PLAYERS // at least one

#if HAVE_LIBAVFILTER
char *vfilters;
char *afilters;
#endif

/* begin core libraries */
#if HAVE_LIBAVUTIL
 #include <libavutil/avstring.h>
 #include <libavutil/mathematics.h>
 #include <libavutil/pixdesc.h>
 #include <libavutil/imgutils.h>
 #include <libavutil/dict.h>
 #include <libavutil/samplefmt.h>
 #include <libavutil/avassert.h>
 #include <libavutil/time.h>
#endif

#if HAVE_LIBAVFORMAT
 #include <libavformat/avformat.h>
#endif

#if HAVE_LIBAVDEVICE
 #include <libavdevice/avdevice.h>
#endif

#if HAVE_LIBSWSCALE
 #include <libswscale/swscale.h>
#endif

#if HAVE_LIBAVUTIL
 #include <libavutil/opt.h> /* AVOptions is requiried to manage bad src image pointers */
#endif

#if HAVE_LIBAVCODEC
 #include <libavcodec/avfft.h>
#endif

#if HAVE_LIBSWRESAMPLE
 #include <libswresample/swresample.h>
#endif
/* end of core libraries */

#if HAVE_LIBAVFILTER
 #include <libavfilter/avcodec.h>
 #include <libavfilter/avfilter.h>
 #include <libavfilter/buffersink.h>
 #include <libavfilter/buffersrc.h>
#endif

#include "fxplayer_opts.h"
#include "fxplayer_queue.h"
#include "fxplayer_clock.h"

#include <assert.h>  // xcb_xlib

#define VIDEO_PICTURE_QUEUE_SIZE        3  /* [-1] <- [0] -> [+1] */

/* SDL audio buffer size, in samples. Should be small to have precise
   Audio/Video sync as SDL does not have hardware buffer fullness info. */
#define SDL_AUDIO_BUFFER_SIZE           1024

static int64_t sws_flags = SWS_BICUBIC;

typedef struct VideoPicture {
  double          pts;             ///< presentation time stamp for this picture
  int64_t         pos;  
#if HAVE_LIBSDL
  SDL_Overlay    *bmp;
#endif
#if HAVE_LIBSDL2
  SDL_Texture    *bmp;
#endif
  int             width; 
  int             height;          ///< source height & width
  int             allocated;
  int             reallocate;
  int             serial;

  AVRational      sample_aspect_ratio;
} VideoPicture;

typedef struct AudioParams {
  int             freq;
  int             channels;
  int64_t         channel_layout;
  enum            AVSampleFormat fmt;
} AudioParams;

typedef struct VideoState {
  SDL_Thread      *decode_tid;         ///< SDL
  SDL_Thread      *video_tid;

  AVInputFormat   *iformat;
  AVFormatContext *avf_ctx;

  int             queue_attachments_req;

  Clock           audclk;              ///< CLOCK
  Clock           vidclk;
  Clock           extclk;

  int             av_sync_type;
  int             realtime;

  double          audio_clock;         ///< AUDIO
  int             audio_clock_serial;
  double          audio_diff_cum;      ///< used for AV difference average computation
  double          audio_diff_avg_coef;
  double          audio_diff_threshold;
  int             audio_diff_avg_count;

  AVStream        *audio_st;
  int             audio_stream;
  int             last_audio_stream; 
  PacketQueue     audioq;

  int             audio_hw_buf_size; 
  uint8_t         silence_buf[SDL_AUDIO_BUFFER_SIZE];
  uint8_t         *audio_buf;
  uint8_t         *audio_buf1;
  unsigned int    audio_buf_size;      ///< in bytes
  unsigned int    audio_buf1_size;
  int             audio_buf_index;     ///< in bytes
  int             audio_buf_frames_pending;

  AVPacket        audio_pkt;
  AVPacket        audio_pkt_tmp;
  int             audio_pkt_tmp_serial;
  int             audio_last_serial;

  struct          AudioParams audio_src;
#if HAVE_LIBAVFILTER
  struct          AudioParams audio_filter_src;
#endif
  struct          AudioParams audio_tgt;

  RDFTContext *rdft;
  int rdft_bits;
  FFTSample *rdft_data;


  double          last_vis_time;      ///< VIDEO

  AVFrame         *frame;
  int             frame_drops_early;
  int             frame_drops_late;
  double          frame_timer;
  double          frame_last_pts;
  double          frame_last_duration;
  double          frame_last_dropped_pts;
  double          frame_last_returned_time;
  double          frame_last_filter_delay;
  int64_t         frame_last_dropped_pos;
  int             frame_last_dropped_serial;
  double          max_frame_duration;      ///< maximum duration of a frame - above this, we consider the jump a timestamp discontinuity 

  AVStream        *video_st;
  int             video_stream;
  int             last_video_stream;
  PacketQueue     videoq;

  VideoPicture    pictq[VIDEO_PICTURE_QUEUE_SIZE];
  int             pictq_size;
  int             pictq_rindex; 
  int             pictq_windex;

  SDL_mutex       *pictq_mutex;
  SDL_cond        *pictq_cond;

  SDL_Rect        last_display_rect;


#if HAVE_LIBAVFILTER
  AVFilterContext *in_video_filter;        ///< the first filter in the video chain
  AVFilterContext *out_video_filter;       ///< the last filter in the video chain
  AVFilterContext *in_audio_filter;        ///< the first filter in the audio chain
  AVFilterContext *out_audio_filter;       ///< the last filter in the audio chain
  AVFilterGraph   *agraph;                 ///< audio filter graph
#endif

  struct          SwrContext *swr_ctx;
#if !HAVE_LIBAVFILTER
  struct          SwsContext *img_convert_ctx;      ///< if libavfilter is not used
#endif

  char            filename[1024];

  int             width;
  int             height;
  int             xleft;
  int             ytop;

  int             pause_curr;           ///< PAUSE
  int             pause_last;
  int             pause_back;

  int             seek_req;             ///< SEEK
  int64_t         seek_pos;
  int64_t         seek_rel;

  int             step;
  int             force_refresh;
  int             quit;

  SDL_cond        *continue_read_thread;

} VideoState;



int configure_filtergraph(AVFilterGraph *graph, const char *filtergraph,
                                 AVFilterContext *source_ctx, AVFilterContext *sink_ctx);


int configure_video_filters(AVFilterGraph *graph, VideoState *vs, const char *vfilters, AVFrame *frame);


int configure_audio_filters(VideoState *vs, const char *afilters, int force_output_format);


#endif /* ENABLE_PLAYERS */

#endif /* FXPLAYER_FILTERS_H */
