/********************************************************************************
*                                                                               *
*                      F X   M O V I E   M A N A G E R                          *
*                                                                               *
*********************************************************************************
* Copyright(C) 2008-2014 by Luigi Ziviani <ziviani_luigi@yahoo.it>.             *
* All Rights Reserved.                                 v 6.6                    *
*********************************************************************************
* Id: fxplayer_main.c,                                 2008/01/01               *
*********************************************************************************
* This file is part of FXMOVIEMANAGER                                           *
*                                                                               *
* FXMOVIEMANAGER is free software; you can redistribute it and/or modify        *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation; either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* FXMOVIEMANAGER is distributed in the hope that it will be useful,             *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with this program; if not, write to the Free Software                   *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA     * 
********************************************************************************/


           /* ****************************************************
            * Acknowledgements:                                  *
            * the code written in this file is based on ffplay:  * 
            * Simple Media Player based on the FFmpeg libraries  *
            * part of FFmpeg, Copyright (c) 2003 Fabrice Bellard *
            *                                                    *
            * I found also useful a tutorial by Stephen Dranger  *
            * who updated the original tutorial by Martin Bohme  *
            **************************************************** */

#include "config.h"

#ifdef ENABLE_PLAYERS
#ifdef HAVE_LIBSDL

#ifdef HAVE_INTTYPES_H
 #include <inttypes.h>
#endif

#include <math.h>
#include <limits.h>
#include <signal.h>
//#include <memory.h>

#include "fxplayer_opts.h"
#include "fxplayer_utils.h"
#include "fxplayer_filters.h"
#include "fxplayer_queue.h"
#include "fxplayer_clock.h"

#include <assert.h>  // xcb_xlib

#define MAX_QUEUE_SIZE                  (15 * 1024 * 1024)
#define MIN_FRAMES                      5

/* no AV sync correction is done if below the minimum AV sync threshold */
#define AV_SYNC_THRESHOLD_MIN           0.01
/* AV sync correction is done if above the maximum AV sync threshold */
#define AV_SYNC_THRESHOLD_MAX           0.1
/* If a frame duration is longer than this, it will not be duplicated to compensate AV sync */
#define AV_SYNC_FRAMEDUP_THRESHOLD      0.1
/* no AV correction is done if too big error */
//#define AV_NOSYNC_THRESHOLD             10.0

/* maximum audio speed change to get correct sync */
#define SAMPLE_CORRECTION_PERCENT_MAX   10
/* we use about AUDIO_DIFF_AVG_NB A-V differences to make the average */
#define AUDIO_DIFF_AVG_NB               20

/* external clock speed adjustment constants for realtime sources based on buffer fullness */
#define EXTERNAL_CLOCK_SPEED_MIN        0.900
#define EXTERNAL_CLOCK_SPEED_MAX        1.010
#define EXTERNAL_CLOCK_SPEED_STEP       0.001
/* polls for possible required screen refresh at least this often, should be less than 1/fps */
#define REFRESH_RATE                    0.01

#define CURSOR_HIDE_DELAY               1000000
/* enum events type and bind to SDL UserEvents*/
#define FF_ALLOC_EVENT                  (SDL_USEREVENT)
#define FF_QUIT_EVENT                   (SDL_USEREVENT + 2)

enum {
  AV_SYNC_AUDIO_MASTER,   ///< default
  AV_SYNC_VIDEO_MASTER,
  AV_SYNC_EXTERNAL_MASTER,
};

static int      av_sync_type = AV_SYNC_AUDIO_MASTER;

static AVInputFormat *iformat = NULL;

static int      cycle_audio_ch = 0;

double          tunerefresh = 0.00;
double          rdftspeed = 0.02;

static int64_t  cursor_last_shown;
static int      cursor_hidden;

static int      time_command = 0;
static int      show_command = 0;

// options passed by the caller
static int64_t  time_s = AV_NOPTS_VALUE;  ///< start time
static int64_t  time_d = AV_NOPTS_VALUE;  ///< duration time
static float    time_a = 0.00;  ///< audio delay time
static int      screen_w = 0;
static int      screen_h = 0;
static int      full = 0;
static int      loop = 0;
static int      drop = 0;
static int      disable_video = 0;
static int      disable_audio = 0;
static int      fs_screen_w = 0;
static int      fs_screen_h = 0;
static int      enlarge_w = 0;
static int      enlarge_h = 0;
static int      display_max = 0;
static int      nolimit_buffer = 0;
static char     *audio_codec_name = NULL;
static char     *video_codec_name = NULL;

static int alsastep = 0;
static int count = 0;
int increment_step = 0;
float increment_size = 0.0;

// the showing surface
static SDL_Surface *screen = NULL;

//static AVPacket flush_pkt;


// AUDIO FORMAT
static inline
int cmp_audio_fmts(enum AVSampleFormat fmt1, int64_t channel_count1,
                   enum AVSampleFormat fmt2, int64_t channel_count2)
{
    /* If channel count == 1, planar and non-planar formats are the same */
    if (channel_count1 == 1 && channel_count2 == 1)
        return av_get_packed_sample_fmt(fmt1) != av_get_packed_sample_fmt(fmt2);
    else
        return channel_count1 != channel_count2 || fmt1 != fmt2;
}

static inline
int64_t get_valid_channel_layout(int64_t channel_layout, int channels)
{
    if (channel_layout && av_get_channel_layout_nb_channels(channel_layout) == channels)
        return channel_layout;
    else
        return 0;
}

// CLOCK
static int get_master_sync_type(VideoState *vs) {

    if (vs->av_sync_type == AV_SYNC_VIDEO_MASTER) {
        if (vs->video_st)
            return AV_SYNC_VIDEO_MASTER;
        else
            return AV_SYNC_AUDIO_MASTER;
    } else if (vs->av_sync_type == AV_SYNC_AUDIO_MASTER) {
        if (vs->audio_st)
            return AV_SYNC_AUDIO_MASTER;
        else
            return AV_SYNC_EXTERNAL_MASTER;
    } else {
        return AV_SYNC_EXTERNAL_MASTER;
    }
}

// get the current master clock value
static double get_master_clock(VideoState *vs) {

    double val;

    switch (get_master_sync_type(vs)) {
        case AV_SYNC_VIDEO_MASTER:
            val = get_clock(&vs->vidclk);
            break;
        case AV_SYNC_AUDIO_MASTER:
            val = get_clock(&vs->audclk);
            break;
        default:
            val = get_clock(&vs->extclk);
            break;
    }
    return val;
}

static void check_external_clock_speed(VideoState *vs) {
   if (vs->video_stream >= 0 && vs->videoq.nb_packets <= MIN_FRAMES / 2 ||
       vs->audio_stream >= 0 && vs->audioq.nb_packets <= MIN_FRAMES / 2) {
       set_clock_speed(&vs->extclk, FFMAX(EXTERNAL_CLOCK_SPEED_MIN, vs->extclk.speed - EXTERNAL_CLOCK_SPEED_STEP));
   } else 
   if ((vs->video_stream < 0 || vs->videoq.nb_packets > MIN_FRAMES * 2) &&
       (vs->audio_stream < 0 || vs->audioq.nb_packets > MIN_FRAMES * 2)) {
       set_clock_speed(&vs->extclk, FFMIN(EXTERNAL_CLOCK_SPEED_MAX, vs->extclk.speed + EXTERNAL_CLOCK_SPEED_STEP));
   } else {
       double speed = vs->extclk.speed;
       if (speed != 1.0)
           set_clock_speed(&vs->extclk, speed + EXTERNAL_CLOCK_SPEED_STEP * (1.0 - speed) / fabs(1.0 - speed));
   }
}

// SEEK
void stream_seek(VideoState *vs, int64_t pos, int64_t rel) {

  if(!vs->seek_req) {
    vs->seek_pos = pos;
    vs->seek_rel = rel;
    vs->seek_req = 1;

    SDL_CondSignal(vs->continue_read_thread);
  }
}

// PAUSE
// toggle pause
static void toggle_pause(VideoState *vs) { 
    
    if (vs->pause_curr) {
        vs->frame_timer += av_gettime() / (double)AV_TIME_BASE + vs->vidclk.pts_drift - vs->vidclk.pts;
        if (vs->pause_back != AVERROR(ENOSYS)) {
            vs->vidclk.paused = 0;
        }
        set_clock(&vs->vidclk, get_clock(&vs->vidclk), vs->vidclk.serial);

    }
    set_clock(&vs->extclk, get_clock(&vs->extclk), vs->extclk.serial);
    vs->pause_curr = vs->audclk.paused = vs->vidclk.paused = vs->extclk.paused = !vs->pause_curr;
}

static void do_pause(VideoState *vs) { 
    toggle_pause(vs);
    vs->step = 0;
}

// if the stream is paused unpause it, then step
static void step_to_next_frame(VideoState *vs) {

    if (vs->pause_curr)
       toggle_pause(vs);
    vs->step = 1;
}

// DISPLAY
// display the command given
static void video_command_display() {
  SDL_Surface *bitmap, *bitmap1;
  SDL_Rect rect, rect1;

  bitmap = SDL_LoadBMP(imgcommand);
  if (bitmap == NULL) {
    fprintf(stderr, "Unable to load %s: %s\n", imgcommand, SDL_GetError());
    return;
  }

  bitmap1 = SDL_LoadBMP(imgtime);
  if (bitmap1 == NULL) {
    fprintf(stderr, "Unable to load %s: %s\n", imgtime, SDL_GetError());
    return;
  }

  rect.x = 0;
  rect.y = 0;
  rect.w = bitmap->w;
  rect.h = bitmap->h;

  rect1.x = 16;
  rect1.y = 0;
  rect1.w = bitmap1->w;
  rect1.h = bitmap1->h;

//  SDL_LockSurface(screen);
  SDL_BlitSurface(bitmap, NULL, screen, &rect);
  SDL_UpdateRects(screen, 1, &rect);
  SDL_BlitSurface(bitmap1, NULL, screen, &rect1);
  SDL_UpdateRects(screen, 1, &rect1);
//  SDL_UnlockSurface(screen);
}

static void calculate_display_rect(SDL_Rect *rect, int scr_xleft, int scr_ytop, int scr_width, int scr_height, VideoPicture *vp) {
    float aspect_ratio;
    int w, h, x, y;

    if (vp->sample_aspect_ratio.num == 0)
        aspect_ratio = 1.0;
    else
        aspect_ratio = av_q2d(vp->sample_aspect_ratio);

    if (aspect_ratio <= 0.0)
        aspect_ratio = 1.0;

    aspect_ratio *= (float)vp->width / (float)vp->height;

    // try and display a picture size keeping a good aspect ratio
    // basing on the assumption that height is the best measure
    h = scr_height;
    w = ((int)rint(h * aspect_ratio)) & ~1;
    // if it fails basing on the assumption that width is the best measure
    if(w > scr_width) {
      w = scr_width;
      h = ((int)rint(w / aspect_ratio)) & ~1;
    }

    // calculate the size of the SDL screen for the picture
    h = video_screen_resize_h(h, enlarge_h);
    w = video_screen_resize_w(w, enlarge_w);
    rect->w = FFMAX(w, 1);
    rect->h = FFMAX(h, 1);
    // calculate the start point coordinates for the picture
    x = (scr_width - w) / 2;
    y = (scr_height - h) / 2;
    rect->x = scr_xleft + x;
    rect->y = scr_ytop  + y;
}


static inline void fill_rectangle(int x, int y, int w, int h, int color) {
    SDL_Rect rect;
    rect.x = x;
    rect.y = y;
    rect.w = w;
    rect.h = h;
    SDL_FillRect(screen, &rect, color);
}

// display the current picture, if any
static void video_image_display(VideoState *vs) {
  VideoPicture *vp;
  AVPicture pict;
  SDL_Rect rect;
  int i;

  vp = &vs->pictq[vs->pictq_rindex];
  if(vp->bmp) {

    calculate_display_rect(&rect, vs->xleft, vs->ytop, vs->width, vs->height, vp);
    // display the window
    SDL_DisplayYUVOverlay(vp->bmp, &rect);

    if (rect.x != vs->last_display_rect.x || 
        rect.y != vs->last_display_rect.y || 
        rect.w != vs->last_display_rect.w || 
        rect.h != vs->last_display_rect.h || 
        vs->force_refresh) {
          int bgcolor = SDL_MapRGB(screen->format, 0x00, 0x00, 0x00);
          fill_rectangle(vs->xleft, vs->ytop, vs->width, vs->height, bgcolor);
          vs->last_display_rect = rect;
    }

    SDL_LockYUVOverlay(vp->bmp);
    SDL_GetClipRect(screen, &rect);

    if (show_command == 1 && time_command <= 10) {
      //fprintf(stderr, "i_timecmd:%i \n", time_command);
      time_command +=1;
      video_command_display();
    } else {
      time_command = 0; 
      show_command = 0;
    }
    SDL_UnlockYUVOverlay(vp->bmp);
  }
}


// FREE MEMORY
static void stream_free(VideoState *vs) {

    VideoPicture *vp;
    int i;
    vs->quit = 1;

    //Wait for the threads to finish
    SDL_WaitThread(vs->decode_tid, NULL);

    packet_queue_destroy(&vs->videoq);
    packet_queue_destroy(&vs->audioq);

    // free the picture filter and bpm buffers
    for(i=0; i<VIDEO_PICTURE_QUEUE_SIZE; i++) {
      vp = &vs->pictq[i];
      if(vp->bmp) {
        SDL_FreeYUVOverlay(vp->bmp);
        vp->bmp = NULL;
      }
    }

    SDL_DestroyMutex(vs->pictq_mutex);
    SDL_DestroyCond(vs->pictq_cond);
    SDL_DestroyCond(vs->continue_read_thread);

#if !HAVE_LIBAVFILTER
    // free the sws scaler context.
    if(vs->img_convert_ctx){
      sws_freeContext(vs->img_convert_ctx);
      vs->img_convert_ctx = NULL;
    }
#endif
    av_free(vs);
}

static void do_exit(VideoState *vs) {

    if(vs) {
      stream_free(vs);
      vs = NULL;
    }
    av_lockmgr_register(NULL);
    uninit_opts();
#ifdef HAVE_LIBAVFILTER
  if(vfilters){
    fprintf(stderr, "player: freeing video filters: %i and audio filters: %i\n", vfilters, afilters);
    vfilters = NULL;
    afilters = NULL;
    av_freep(&vfilters);
    av_freep(&afilters);
  }
#endif
    // Close the network protocol
    avformat_network_deinit();

    SDL_Quit();

    av_log(NULL, AV_LOG_QUIET, "%s\n", "Thank you for using FXMovieManager");
    return;
}

// VIDEO
// set the SDL screen
static int video_screen_open(VideoState *vs, int force_mode, VideoPicture *vp){

    int sw, sh, sdl_flags, sdl_bpp;
    SDL_Rect rect;

    if(full==1) sdl_flags = SDL_HWSURFACE|SDL_ASYNCBLIT|SDL_HWACCEL|SDL_FULLSCREEN;
    else        sdl_flags = SDL_HWSURFACE|SDL_ASYNCBLIT|SDL_HWACCEL|SDL_RESIZABLE;

    if (full==1 && fs_screen_w) {
        sw = fs_screen_w;
        sh = fs_screen_h;
    } else if(full==0 && screen_w){
        sw = screen_w;
        sh = screen_h;
    } else if(vp && vp->width){
        calculate_display_rect(&rect, 0, 0, INT_MAX, vp->height, vp);
        if(increment_step){
           increment_size = screen_increment_conversion(increment_size, increment_step); // GUI command
           sw = increment_size * rect.w;
           sh = increment_size * rect.h;
           av_log(NULL, AV_LOG_INFO, "GUI video_screen_open: %d: %d:\n", sw, sh);
        }else{
           sw = rect.w;
           sh = rect.h;
           av_log(NULL, AV_LOG_INFO, "SDL video_screen_open: %d: %d:\n", sw, sh);
        }
    } else {
        sw = 720;
        sh = 400;
    }
    sw = FFMIN(16383, sw);

    if(screen
       && vs->width  == screen->w && screen->w  == sw 
       && vs->height == screen->h && screen->h  == sh 
       && !force_mode)
          return 0; //nothing to do

    // set the window icon
    SDL_Surface *bitmap;
    bitmap = SDL_LoadBMP(ICON_LG);
    if (bitmap == NULL) {
       av_log(NULL, AV_LOG_INFO, "Unable to load %s: %s\n", ICON_LG, SDL_GetError());
       return;
    }
    SDL_WM_SetIcon(bitmap,NULL);

#if HAVE_STRUCT_SDL_VIDEOINFOP
    const SDL_VideoInfo *sdl_vi = SDL_GetVideoInfo();
    switch (sdl_vi->vfmt->BitsPerPixel) {
       case 16:
       case 24:
       case 32:
          sdl_bpp = sdl_vi->vfmt->BitsPerPixel;
          break;
       default:
          sdl_bpp = 0;
          break;
    }
#else
      sdl_bpp = 0;
#endif
    // set the SDL screen
    screen = SDL_SetVideoMode(sw, sh, sdl_bpp, sdl_flags);
    if (!screen) {
        av_log(NULL, AV_LOG_FATAL, "SDL: Unable to set video mode -> quitting\n");
        do_exit(vs);
    }
    //set the window caption
    char cap_title[1024]="FXMM: ";
    strncat(cap_title, vs->filename, sizeof(cap_title));
    SDL_WM_SetCaption( cap_title, NULL );
    //SDL_WM_SetCaption( cap_title, cap_title );
    // store the VideoState screen size (same of SDL screen size)
    vs->width  = screen->w;
    vs->height = screen->h;

    return 0;
}

// call subroutines for video display
static void video_display(VideoState *vs) {
    if(!screen) {
       av_log(NULL, AV_LOG_INFO, "INIT: video_screen_open\n");
       video_screen_open(vs, 0, NULL);
    }
    if(vs->video_st) {
       //av_log(NULL, AV_LOG_INFO, "INIT: video_image_display\n");
       video_image_display(vs);
    } else if (vs->audio_st) {
       av_log(NULL, AV_LOG_INFO, "INIT: filling video rectangle\n");
       int fgcolor = SDL_MapRGB(screen->format, 0xff, 0xff, 0xff);
       fill_rectangle(0, 0, screen->w, screen->h, fgcolor);
    }
}



// synchronize the video stream
// return a new video buffer size (if found AUDIO or EXTERNAL MASTER)
// deleting or adding samples to better sync with master clock
static double synchronize_video(double delay, VideoState *vs) {

    double sync_threshold, diff_clock;

    // if VIDEO is not the master clock, then we try 
    // to remove or duplicate samples from VIDEO stream
    // in order to correct the clock according to the AUDIO master
    if (get_master_sync_type(vs) != AV_SYNC_VIDEO_MASTER) {
        // cumuate
        diff_clock = get_clock(&vs->vidclk) - get_master_clock(vs);

        sync_threshold = FFMAX(AV_SYNC_THRESHOLD_MIN, FFMIN(AV_SYNC_THRESHOLD_MAX, delay));
        //av_log(NULL, AV_LOG_INFO, "sync threshold=%0.3f\n", sync_threshold); // DEBUG

        // SKIP or REPEAT video frame if in the used range
        if (!isnan(diff_clock) && fabs(diff_clock) < vs->max_frame_duration) {

            if (diff_clock <= -sync_threshold)
                delay = FFMAX(0, delay + diff_clock);
            else if (diff_clock >= sync_threshold && delay > AV_SYNC_FRAMEDUP_THRESHOLD)
                delay = delay + diff_clock;
            else if (diff_clock >= sync_threshold);
                delay = 2 * delay; 
        }

        // in case the difference is too big for automatic video synchro, 
        // we can use the AUDIO_DELAY option to correct this difference.
        if(time_a != 0) {
            delay -= (double)(time_a);
            //av_log(NULL, AV_LOG_INFO, "manual: delay=%0.3f\n", delay); // DEBUG
        }
    }

    //av_log(NULL, AV_LOG_INFO, "video: delay=%0.3f A-V=%f\n", delay, -diff_clock); // DEBUG

    return delay;
}

// update queue size and signal for next picture
static void pictq_next_picture(VideoState *vs) {
    //printf("next picture rindex=%i\n", vs->pictq_rindex);
    if (++vs->pictq_rindex == VIDEO_PICTURE_QUEUE_SIZE)
       vs->pictq_rindex = 0;

    SDL_LockMutex(vs->pictq_mutex);
    vs->pictq_size--;
    SDL_CondSignal(vs->pictq_cond);
    SDL_UnlockMutex(vs->pictq_mutex);
}

// update queue size and signal for previews picture
static int pictq_prev_picture(VideoState *vs) {
    //printf("prev picture rindex=%i\n", vs->pictq_rindex);
    VideoPicture *prevvp;
    int ret = 0;
    prevvp = &vs->pictq[(vs->pictq_rindex + VIDEO_PICTURE_QUEUE_SIZE - 1) % VIDEO_PICTURE_QUEUE_SIZE];
    if (prevvp->allocated && prevvp->serial == vs->videoq.serial) {
        SDL_LockMutex(vs->pictq_mutex);

        if (vs->pictq_size < VIDEO_PICTURE_QUEUE_SIZE) {
            if (--vs->pictq_rindex == -1)
                vs->pictq_rindex = VIDEO_PICTURE_QUEUE_SIZE - 1;
            vs->pictq_size++;
            ret = 1;
        }
        SDL_CondSignal(vs->pictq_cond);
        SDL_UnlockMutex(vs->pictq_mutex);
    }
    return ret;
}

// update current video pts
static void update_video_pts(VideoState *vs, double pts, int64_t pos, int serial) {

    set_clock(&vs->vidclk, pts, serial);
    sync_clock_to_slave(&vs->extclk, &vs->vidclk);
    vs->frame_last_pts = pts;
}

// called to display each frame
static void video_refresh_timer(void *opaque, double *remaining_time) {

  VideoState *vs = opaque;
  VideoPicture *vp;
  double time_curr;

  if (!vs->pause_curr && get_master_sync_type(vs) == AV_SYNC_EXTERNAL_MASTER && vs->realtime)
     check_external_clock_speed(vs);

  if (disable_video==1 && vs->audio_st) { // DEBUG
      time_curr = av_gettime() / (double)AV_TIME_BASE;
      if (vs->force_refresh || vs->last_vis_time + rdftspeed < time_curr) {
          video_display(vs);
          vs->last_vis_time = time_curr;
      }
      *remaining_time = FFMIN(*remaining_time, vs->last_vis_time + rdftspeed - time_curr);
  }

  // VIDEO STREAM 
  if(vs->video_st) {
     int redisplay = 0;
     if (vs->force_refresh)
        redisplay = pictq_prev_picture(vs);
tryagain:
     if(vs->pictq_size == 0 ) {
        //printf("picture queue size is = 0\n");
        SDL_LockMutex(vs->pictq_mutex);
        if(vs->frame_last_dropped_pts != AV_NOPTS_VALUE && vs->frame_last_dropped_pts > vs->frame_last_pts) {
           update_video_pts(vs, vs->frame_last_dropped_pts, vs->frame_last_dropped_pos, vs->frame_last_dropped_serial);
           vs->frame_last_dropped_pts =  AV_NOPTS_VALUE;
        }
        SDL_UnlockMutex(vs->pictq_mutex);
       // nothing to do, no picture to display in the queue
     } else {
        double last_duration, duration, delay;
        // dequeue the picture
        vp = &vs->pictq[vs->pictq_rindex];

        if (vp->serial != vs->videoq.serial) {
           av_log(NULL, AV_LOG_INFO, "request next vp=%i, videoq=%i\n", vp->serial, vs->videoq.serial);
           pictq_next_picture(vs);
           redisplay = 0;
           goto tryagain;
        }

        if (vs->pause_curr)
           goto display;

        // compute nominal last_duration
        last_duration = vp->pts - vs->frame_last_pts;
        // if duration of the last frame was sane, update last_duration in video state
        if (!isnan(last_duration) && last_duration > 0 && last_duration < vs->max_frame_duration) {
           vs->frame_last_duration = last_duration;
        }
        if (redisplay)
            delay = 0.0;
        else // call to sync the video stream with the master one
            delay = synchronize_video(vs->frame_last_duration, vs);

// -----DEBUG frame timer ------
        time_curr = av_gettime() / (double)AV_TIME_BASE;

        if(time_curr < vs->frame_timer + delay && !redisplay) {
            *remaining_time = FFMIN(vs->frame_timer + delay - time_curr, *remaining_time);
            //av_log(NULL, AV_LOG_INFO, 
            //       "remaining_time=%0.3f delay=%0.3f  time_curr=%0.3f frame_timer=%0.3f, \n", 
            //        *remaining_time,     delay,       time_curr,      vs->frame_timer);
            return; // FIXME
        }

        vs->frame_timer += delay;
        //av_log(NULL, AV_LOG_INFO, "new frame_timer=%0.3f\n", vs->frame_timer);

        if (delay > 0 && time_curr - vs->frame_timer > AV_SYNC_THRESHOLD_MAX)
            vs->frame_timer = time_curr;

        SDL_LockMutex(vs->pictq_mutex);
        if (!redisplay && !isnan(vp->pts))
            update_video_pts(vs, vp->pts, vp->pos, vp->serial);
        SDL_UnlockMutex(vs->pictq_mutex);

        if(vs->pictq_size > 1) {
            VideoPicture *nextvp = &vs->pictq[(vs->pictq_rindex + 1) % VIDEO_PICTURE_QUEUE_SIZE];
            duration = nextvp->pts - vp->pts; 
            // if video frame drop
            if(!vs->step && (redisplay || drop==1 || (drop==0 && get_master_sync_type(vs) != AV_SYNC_VIDEO_MASTER)) && time_curr > vs->frame_timer + duration){
               if (!redisplay)
                  vs->frame_drops_late++;
               pictq_next_picture(vs);
               redisplay = 0;
               goto tryagain;
            } 
        }

display:
      // show the picture
      if(disable_video == 0)
         video_display(vs);

      pictq_next_picture(vs);

      if(vs->step && !vs->pause_curr)
         toggle_pause(vs);
    }

  // AUDIO STREAM  -- DEBUG??
  } else if (vs->audio_st) {
     // display a rectangle
     if (disable_video == 0)
        video_display(vs);
  }

  vs->force_refresh = 0;
}


// PICTURES 
// allocate a pictude: it needs to be done in main thread to avoid potential locking problems
static void alloc_picture(VideoState *vs) {

  VideoPicture *vp;
  int64_t diff_buffer;

  vp = &vs->pictq[vs->pictq_windex];

  if(vp->bmp)
    SDL_FreeYUVOverlay(vp->bmp);

  video_screen_open(vs, 0, vp);

  // Allocate a place to put our YUV image on that screen
  vp->bmp = SDL_CreateYUVOverlay(vp->width,
                                 vp->height,
                                 SDL_YV12_OVERLAY,
                                 screen);

  diff_buffer = vp->bmp ? FFMAX(vp->bmp->pixels[0], vp->bmp->pixels[1]) - FFMIN(vp->bmp->pixels[0], vp->bmp->pixels[1]) : 0;
  /* SDL allocates a buffer smaller than requested if the video
   * overlay hardware is unable to support the requested size. */
  if (!vp->bmp || vp->bmp->pitches[0] < vp->width || diff_buffer < (int64_t)vp->height * vp->bmp->pitches[0]) {
       av_log(NULL, AV_LOG_FATAL,
              "Error: the video system does not support an image size of %dx%d pixels\n", vp->width, vp->height );
    do_exit(vs);
  }
  
  SDL_LockMutex(vs->pictq_mutex);
  vp->allocated = 1;
  SDL_CondSignal(vs->pictq_cond);
  SDL_UnlockMutex(vs->pictq_mutex);
}

static void duplicate_right_border_pixels(SDL_Overlay *bmp) {
    int i, width, height;
    Uint8 *p, *maxp;
    for (i = 0; i < 3; i++) {
        width  = bmp->w;
        height = bmp->h;
        if (i > 0) {
            width  >>= 1;
            height >>= 1;
        }
        if (bmp->pitches[i] > width) {
            maxp = bmp->pixels[i] + bmp->pitches[i] * height - 1;
            for (p = bmp->pixels[i] + width - 1; p < maxp; p += bmp->pitches[i])
                *(p+1) = *p;
        }
    }
}

static int queue_picture(VideoState *vs, AVFrame *src_frame, double pts, int64_t pos, int serial) {

  VideoPicture *vp;
  //printf("frame_type=%c pts=%0.3f\n", av_get_picture_type_char(src_frame->pict_type), pts);

  // wait until there is space for a new picture
  SDL_LockMutex(vs->pictq_mutex);
  // keep the last already displayed picture in the queue
  while(vs->pictq_size >= VIDEO_PICTURE_QUEUE_SIZE - 1 && !vs->videoq.quit) {
     SDL_CondWait(vs->pictq_cond, vs->pictq_mutex);
  }
  SDL_UnlockMutex(vs->pictq_mutex);

  if(vs->videoq.quit)
     return -1;

  // windex is set to 0 initially
  vp = &vs->pictq[vs->pictq_windex];
  vp->sample_aspect_ratio = src_frame->sample_aspect_ratio;

  // if NOT FRAME: allocate or resize hardware picture buffer
  if(!vp->bmp || vp->reallocate || !vp->allocated ||
     vp->width  != src_frame->width ||
     vp->height != src_frame->height) {

     vp->allocated = 0;
     vp->reallocate = 0;
     vp->width = src_frame->width;
     vp->height = src_frame->height;

     // the allocation must be done in the main thread to avoid locking problems
     SDL_Event event;
     event.type = FF_ALLOC_EVENT;
     event.user.data1 = vs;
     SDL_PushEvent(&event);

     // wait until there is a picture allocated
     SDL_LockMutex(vs->pictq_mutex);
     while(!vp->allocated && !vs->videoq.quit) {
        SDL_CondWait(vs->pictq_cond, vs->pictq_mutex);
     }
     // if the queue is aborted, we have to pop the pending ALLOC event or wait for the allocation to complete 
     if(vs->videoq.quit && SDL_PeepEvents(&event, 1, SDL_GETEVENT, SDL_EVENTMASK(FF_ALLOC_EVENT)) != 1) {
        while (!vp->allocated) {
           SDL_CondWait(vs->pictq_cond, vs->pictq_mutex);
        }
     }
     SDL_UnlockMutex(vs->pictq_mutex);

     if(vs->videoq.quit)
        return -1;
  }

  // if FRAME: if the frame is not skipped display it
  if(vp->bmp) {
     AVPicture pict = { { 0 } };

     // get a pointer to the bitmap
     SDL_LockYUVOverlay(vp->bmp);

     pict.data[0] = vp->bmp->pixels[0];
     pict.data[1] = vp->bmp->pixels[2];
     pict.data[2] = vp->bmp->pixels[1];
    
     pict.linesize[0] = vp->bmp->pitches[0];
     pict.linesize[1] = vp->bmp->pitches[2];
     pict.linesize[2] = vp->bmp->pitches[1];

#if HAVE_LIBAVFILTER
     // DIRECT RENDERING (FIXME: no pix format conversion)
     av_picture_copy(&pict, 
                     (AVPicture *)src_frame,
                     src_frame->format, 
                     vp->width, 
                     vp->height);
#else
     av_opt_get_int(sws_opts, "sws_flags", 0, &sws_flags);
     // Checks if sws context can be reused, otherwise reallocates a new one.
     vs->img_convert_ctx = sws_getCachedContext(vs->img_convert_ctx,
                                                vp->width, vp->height, src_frame->format,// source format
                                                vp->width, vp->height, PIX_FMT_YUV420P,  // destination in planar YUV
                                                sws_flags, 
                                                NULL, NULL,                              // img filters
                                                NULL);
     // If sws context is NULL
     if(vs->img_convert_ctx == NULL) {
        av_log(NULL, AV_LOG_FATAL, "Cannot initialize the conversion context\n");
	    return 1;
     }

     // Scales the image slice in src_Slice and puts the resulting scaled
     // slice in the image in dst (a slice is a sequence of consecutive rows in an image)
     sws_scale(vs->img_convert_ctx, 
              (const uint8_t * const*)src_frame->data, 
              src_frame->linesize,
              0, 
              vp->height, 
              pict.data, 
              pict.linesize);
#endif
     // workaround SDL PITCH_WORKAROUND
     duplicate_right_border_pixels(vp->bmp);
     // pdate the bitmap content 
     SDL_UnlockYUVOverlay(vp->bmp);

     vp->pts = pts;
     vp->pos = pos;
     vp->serial = serial;

     // at last update the picture_windex (count)
     if(++vs->pictq_windex == VIDEO_PICTURE_QUEUE_SIZE)
        vs->pictq_windex = 0;

     SDL_LockMutex(vs->pictq_mutex);
     vs->pictq_size++;
     SDL_UnlockMutex(vs->pictq_mutex);
  }
  return 0;
}

static int get_video_frame(VideoState *vs, AVFrame *frame, AVPacket *pkt, int *serial) {

    int got_picture;

    if (packet_queue_get(&vs->videoq, pkt, 1, serial) < 0)
        return -1;

    if (pkt->data == flush_pkt.data) {
        avcodec_flush_buffers(vs->video_st->codec);

        SDL_LockMutex(vs->pictq_mutex);
        // Make sure there are no long delay timers 
        // ideally should just flush the queue but that's harder
        while (vs->pictq_size && !vs->videoq.quit) {
            SDL_CondWait(vs->pictq_cond, vs->pictq_mutex);
        }
        vs->frame_last_pts = AV_NOPTS_VALUE;
        vs->frame_last_duration = 0;
        vs->frame_timer = (double)(av_gettime() / (double)AV_TIME_BASE);
        vs->frame_last_dropped_pts = AV_NOPTS_VALUE;
        SDL_UnlockMutex(vs->pictq_mutex);

        return 0;
    }

    if(avcodec_decode_video2(vs->video_st->codec, frame, &got_picture, pkt) < 0) 
        return 0;

    if (got_picture) {
        int ret = 1;  // set the error TRUE
        double dpts = NAN;

        frame->pts = av_frame_get_best_effort_timestamp(frame);
        if (frame->pts == AV_NOPTS_VALUE)
           frame->pts = frame->pkt_dts;
        if (frame->pts != AV_NOPTS_VALUE)
            dpts = av_q2d(vs->video_st->time_base) * frame->pts;

        frame->sample_aspect_ratio = av_guess_sample_aspect_ratio(vs->avf_ctx, vs->video_st, frame);

        if (drop==1 || (drop==0 && get_master_sync_type(vs) != AV_SYNC_VIDEO_MASTER)) {
            SDL_LockMutex(vs->pictq_mutex);
            if (vs->frame_last_pts != AV_NOPTS_VALUE && frame->pts != AV_NOPTS_VALUE) {
                double diff_clock = get_clock(&vs->vidclk) - get_master_clock(vs);
                double diff_pts = dpts - vs->frame_last_pts;
                if (!isnan(diff_clock) && fabs(diff_clock) < AV_NOSYNC_THRESHOLD &&
                    !isnan(diff_pts) && diff_pts > 0 && diff_pts < AV_NOSYNC_THRESHOLD &&
                    diff_clock + diff_pts - vs->frame_last_filter_delay < 0 &&
                    vs->videoq.nb_packets) {
                      vs->frame_last_dropped_pos = pkt->pos;
                      vs->frame_last_dropped_pts = dpts;
                      vs->frame_last_dropped_serial = *serial;
                      vs->frame_drops_early++;
                      av_frame_unref(frame);
                      ret = 0; // error is FALSE
                }
            }
            SDL_UnlockMutex(vs->pictq_mutex);
        } 

        return ret;
    }
    return 0;
}

static int video_thread(void *arg) {

  AVPacket pkt = { 0 };
  VideoState *vs = arg;
  AVFrame *p_frame = av_frame_alloc();
  double pts;
  int ret;
  int serial = 0;

#if HAVE_LIBAVFILTER
    AVFilterGraph *graph = avfilter_graph_alloc();
    AVFilterContext *filt_out = NULL, *filt_in = NULL;
    int last_w = 0;
    int last_h = 0;
    enum AVPixelFormat last_format = -2;
    int last_serial = -1;
#endif

  for(;;) {
    while (vs->pause_curr && !vs->videoq.quit)
       SDL_Delay(10);

    avcodec_get_frame_defaults(p_frame);
    av_free_packet(&pkt);

    ret = get_video_frame(vs, p_frame, &pkt, &serial); 
    if (ret < 0)
       goto the_end;
    if (!ret)
       continue;

#if HAVE_LIBAVFILTER
        if (   last_w      != p_frame->width
            || last_h      != p_frame->height
            || last_format != p_frame->format
            || last_serial != serial) { 
            av_log(NULL, AV_LOG_INFO, "Video frame changed from size:%dx%d format:%s serial:%d to size:%dx%d format:%s serial:%d\n",
                   last_w, last_h,
                   (const char *)av_x_if_null(av_get_pix_fmt_name(last_format), "none"), last_serial,
                   p_frame->width, p_frame->height,
                   (const char *)av_x_if_null(av_get_pix_fmt_name(p_frame->format), "none"), serial);
           avfilter_graph_free(&graph);
           graph = avfilter_graph_alloc();
           if ((ret = configure_video_filters(graph, vs, vfilters, p_frame)) < 0) {
               SDL_Event event;
               event.type = FF_QUIT_EVENT;
               event.user.data1 = vs;
               SDL_PushEvent(&event);
               av_free_packet(&pkt);
               goto the_end;
           }
           filt_out    = vs->out_video_filter;
           filt_in     = vs->in_video_filter;
           last_w      = p_frame->width;
           last_h      = p_frame->height;
           last_format = p_frame->format;
           last_serial = serial;
        }

        ret = av_buffersrc_add_frame(filt_in, p_frame);
        if (ret < 0)
            goto the_end;

        av_frame_unref(p_frame);
        avcodec_get_frame_defaults(p_frame);
        av_free_packet(&pkt);

        while (ret >= 0) {
            vs->frame_last_returned_time = av_gettime() / (double)AV_TIME_BASE;

            ret = av_buffersink_get_frame_flags(filt_out, p_frame, 0);
            if (ret < 0) {
                ret = 0;
                break;
            }

            vs->frame_last_filter_delay = av_gettime() / (double)AV_TIME_BASE - vs->frame_last_returned_time;
            if (fabs(vs->frame_last_filter_delay) > AV_NOSYNC_THRESHOLD / 10.0)
                vs->frame_last_filter_delay = 0;

            pts = (p_frame->pts == AV_NOPTS_VALUE) ? NAN : p_frame->pts * av_q2d(filt_out->inputs[0]->time_base);
            ret = queue_picture(vs, p_frame, pts, av_frame_get_pkt_pos(p_frame), serial);
            av_frame_unref(p_frame);
        }
#else
        pts = (p_frame->pts == AV_NOPTS_VALUE) ? NAN : p_frame->pts * av_q2d(vs->video_st->time_base);
        ret = queue_picture(vs, p_frame, pts, pkt.pos, serial);
        av_frame_unref(p_frame);
#endif

       if (ret < 0) 
          goto the_end;
    }
 the_end:
    avcodec_flush_buffers(vs->video_st->codec);
#if HAVE_LIBAVFILTER
    avfilter_graph_free(&graph);
#endif
    av_free_packet(&pkt);
    av_frame_free(&p_frame);
    return 0;
}


// synchronize the audio stream:
// return a new audio buffer size (if found video or external master)
// deleting or adding samples to better sync with master clock
// and usefull also for passing an audiodelay value to the player
static int synchronize_audio(VideoState *vs, int nb_samples) {

  int delay_nb_samples; 
  int wanted_nb_samples = nb_samples;

  // if AUDIO is not the master clock, then we try to remove or add samples from AUDIO
  // in order to correct the clock according to the VIDEO master
  if (get_master_sync_type(vs) != AV_SYNC_AUDIO_MASTER) {
    double diff_clock, avg_diff;
    int min_nb_samples, max_nb_samples;

    // cumuate till we have a Master > Audio (every time base)
    diff_clock =  get_clock(&vs->audclk) - get_master_clock(vs);

	// if the threshold is not reached 
    if (!isnan(diff_clock) && fabs(diff_clock) < AV_NOSYNC_THRESHOLD) {
      // cumulate the diffs
      vs->audio_diff_cum = diff_clock + vs->audio_diff_avg_coef * vs->audio_diff_cum;

      if(vs->audio_diff_avg_count < AUDIO_DIFF_AVG_NB) {
        // not enough measures to have a correct estimate
	    vs->audio_diff_avg_count++;
      } else {
        // estimate the A-V difference
        avg_diff = vs->audio_diff_cum * (1.0 - vs->audio_diff_avg_coef);

        if(fabs(avg_diff) >= vs->audio_diff_threshold) {
	      wanted_nb_samples = nb_samples + (int)(diff_clock * vs->audio_src.freq);
          // in case the calculation fails and audio adjustment is needed, 
          // we manually increase or decrease the wanted size 
          // so that the real sample size will be appended or cutted
          if(time_a !=0) {
            delay_nb_samples = (int)(time_a * vs->audio_src.freq);
            wanted_nb_samples = nb_samples + delay_nb_samples;
          } else {
            // CALCULATION: estimate the A-V difference
            min_nb_samples = ((nb_samples * (100 - SAMPLE_CORRECTION_PERCENT_MAX) / 100));
            max_nb_samples = ((nb_samples * (100 + SAMPLE_CORRECTION_PERCENT_MAX) / 100));
            wanted_nb_samples = FFMIN(FFMAX(wanted_nb_samples, min_nb_samples), max_nb_samples);
	      } 
        } 
      }
    } else {
    // the difference is too big; RESET A-V filter to zero
    vs->audio_diff_avg_count = 0;
    vs->audio_diff_cum = 0;
    }
  }
  return wanted_nb_samples;
}

// decode one audio frame and return its uncompressed size
// the bites of the processed audio frame is stored in vs->audio_buf
static int audio_decode_frame(VideoState *vs) {

  AVPacket *pkt = &vs->audio_pkt;
  AVPacket *pkt_tmp = &vs->audio_pkt_tmp;
  AVCodecContext *dec = vs->audio_st->codec;
  int64_t dec_channel_layout;
  int or_data_size, cv_data_size, len_a, len_b;
  int new_pkt = 0;
  int flush_done = 0;
  int got_frame;
  int wanted_nb_samples;
  AVRational tb;
  int ret;
  int reconfigure;
  //av_unused double audio_clock0;

  for(;;) {
    // an audio packet can contain several frames
    while(pkt_tmp->size > 0 || (!pkt_tmp->data && new_pkt) || vs->audio_buf_frames_pending){
      if (!vs->frame) {
        if (!(vs->frame = av_frame_alloc()))
          return AVERROR(ENOMEM);
      } else
        av_frame_unref(vs->frame);
        avcodec_get_frame_defaults(vs->frame);

      if (vs->audioq.serial != vs->audio_pkt_tmp_serial)
        break;

      if (vs->pause_curr)
        return -1;

      if (!vs->audio_buf_frames_pending) {
        if(flush_done)
          break;

        new_pkt = 0;
        len_a = avcodec_decode_audio4(dec, vs->frame, &got_frame, pkt_tmp); 
        // if error, skip the frame
        if(len_a < 0) {
	      pkt_tmp->size = 0;
	      break;
        }

        pkt_tmp->data += len_a;  // increase
        pkt_tmp->size -= len_a;  // decrease

        if(!got_frame) {
          // stop sending empty packets because the decoding is finished
           if (!pkt_tmp->data && dec->codec->capabilities & CODEC_CAP_DELAY)
              flush_done = 1;
           // No data yet, get next frame
	       continue;
        }

        tb = (AVRational){1, vs->frame->sample_rate};

        if(vs->frame->pts != AV_NOPTS_VALUE) // DEBUG (0, AV_NOPTS_VALUE)
           vs->frame->pts = av_rescale_q(vs->frame->pts, dec->time_base, tb);
        else 
        if(vs->frame->pkt_pts != AV_NOPTS_VALUE) 
           vs->frame->pts = av_rescale_q(vs->frame->pkt_pts, vs->audio_st->time_base, tb);

        if(pkt_tmp->pts != AV_NOPTS_VALUE)
           pkt_tmp->pts += (double)vs->frame->nb_samples / vs->frame->sample_rate / av_q2d(vs->audio_st->time_base);



#if HAVE_LIBAVFILTER
                dec_channel_layout = get_valid_channel_layout(vs->frame->channel_layout, av_frame_get_channels(vs->frame));

                reconfigure =
                    cmp_audio_fmts(vs->audio_filter_src.fmt, vs->audio_filter_src.channels,
                                   vs->frame->format, av_frame_get_channels(vs->frame))    ||
                    vs->audio_filter_src.channel_layout != dec_channel_layout ||
                    vs->audio_filter_src.freq           != vs->frame->sample_rate ||
                    vs->audio_pkt_tmp_serial            != vs->audio_last_serial;

                if (reconfigure) {
                    char buf1[1024], buf2[1024];
                    av_get_channel_layout_string(buf1, sizeof(buf1), -1, vs->audio_filter_src.channel_layout);
                    av_get_channel_layout_string(buf2, sizeof(buf2), -1, dec_channel_layout);
                    av_log(NULL, AV_LOG_DEBUG,
                           "Audio frame changed from rate:%d ch:%d fmt:%s layout:%s serial:%d to rate:%d ch:%d fmt:%s layout:%s serial:%d\n",
                           vs->audio_filter_src.freq, vs->audio_filter_src.channels, av_get_sample_fmt_name(vs->audio_filter_src.fmt), buf1, vs->audio_last_serial,
                           vs->frame->sample_rate, av_frame_get_channels(vs->frame), av_get_sample_fmt_name(vs->frame->format), buf2, vs->audio_pkt_tmp_serial);

                    vs->audio_filter_src.fmt            = vs->frame->format;
                    vs->audio_filter_src.channels       = av_frame_get_channels(vs->frame);
                    vs->audio_filter_src.channel_layout = dec_channel_layout;
                    vs->audio_filter_src.freq           = vs->frame->sample_rate;
                    vs->audio_last_serial               = vs->audio_pkt_tmp_serial;

                    if ((ret = configure_audio_filters(vs, afilters, 1)) < 0)
                        return ret;
                }

                if ((ret = av_buffersrc_add_frame(vs->in_audio_filter, vs->frame)) < 0)
                    return ret;
                av_frame_unref(vs->frame);
#endif
            }
#if HAVE_LIBAVFILTER
            if ((ret = av_buffersink_get_frame_flags(vs->out_audio_filter, vs->frame, 0)) < 0) {
                if (ret == AVERROR(EAGAIN)) {
                    vs->audio_buf_frames_pending = 0;
                    continue;
                }
                return ret;
            }
            vs->audio_buf_frames_pending = 1;
            tb = vs->out_audio_filter->inputs[0]->time_base;
#endif

      // original audio data size
      or_data_size = av_samples_get_buffer_size(NULL, 
                                                av_frame_get_channels(vs->frame),
                                                vs->frame->nb_samples, 
                                                vs->frame->format, 
                                                1);

      dec_channel_layout = (vs->frame->channel_layout && av_frame_get_channels(vs->frame) == av_get_channel_layout_nb_channels(vs->frame->channel_layout)) ? 
                            vs->frame->channel_layout : 
                            av_get_default_channel_layout(av_frame_get_channels(vs->frame)); 

      // audio synchronization
      wanted_nb_samples = synchronize_audio(vs, vs->frame->nb_samples);

      // if conversion has not be done yet try to convert
      if (dec_channel_layout     != vs->audio_src.channel_layout || 
          vs->frame->format      != vs->audio_src.fmt            || 
          vs->frame->sample_rate != vs->audio_src.freq           ||
          (wanted_nb_samples     != vs->frame->nb_samples && !vs->swr_ctx)){
        // free the converter buffer
        swr_free(&vs->swr_ctx);
        // try and use the audio converter
        vs->swr_ctx = swr_alloc_set_opts(NULL,
                                         vs->audio_tgt.channel_layout, vs->audio_tgt.fmt, vs->audio_tgt.freq,
                                         dec_channel_layout,           vs->frame->format, vs->frame->sample_rate,
                                         0, 
                                         NULL);

        // if unable to convert audio skip setting converted variables and print to stdout
        if (!vs->swr_ctx || swr_init(vs->swr_ctx) < 0) {
          av_log(NULL, AV_LOG_ERROR,
                 "Unable to convert FROM sample rate %d Hz %s format %d channels TO sample rate %d Hz %s format %d channels!\n",
          vs->frame->sample_rate,
          av_get_sample_fmt_name(vs->frame->format),
          av_frame_get_channels(vs->frame),
          vs->audio_tgt.freq,
          av_get_sample_fmt_name(vs->audio_tgt.fmt),
          vs->audio_tgt.channels);
          break; // exit the loop
        }
        // set converted variables
        vs->audio_src.channel_layout = dec_channel_layout;
        vs->audio_src.fmt = vs->frame->format;
        vs->audio_src.freq = vs->frame->sample_rate;
        vs->audio_src.channels = av_frame_get_channels(vs->frame);
      }

      // if conversion was done
      if (vs->swr_ctx) {
        uint8_t **out = &vs->audio_buf1;
        int out_count = (int64_t)wanted_nb_samples * vs->audio_tgt.freq / vs->frame->sample_rate + 256;
        int out_size  = av_samples_get_buffer_size(NULL, vs->audio_tgt.channels, out_count, vs->audio_tgt.fmt, 0);
        const uint8_t **in = (const uint8_t **) vs->frame->extended_data;

        if (out_size < 0) {
          av_log(NULL, AV_LOG_ERROR, "av_samples_get_buffer_size() failed\n");
          break;
        }

        if (wanted_nb_samples != vs->frame->nb_samples) {
          if (swr_set_compensation(vs->swr_ctx, (wanted_nb_samples - vs->frame->nb_samples) * vs->audio_tgt.freq / vs->frame->sample_rate,
                                   wanted_nb_samples * vs->audio_tgt.freq / vs->frame->sample_rate) < 0) {
            av_log(NULL, AV_LOG_ERROR, "Audio compensation failed\n");
            break;
          }
        }

        av_fast_malloc(&vs->audio_buf1, &vs->audio_buf1_size, out_size);
        if (!vs->audio_buf1)
          return AVERROR(ENOMEM);

        // calculate converted audio buffer
        len_b = swr_convert(vs->swr_ctx, 
                            out, 
                            out_count,
                            in, 
                            vs->frame->nb_samples);

        if (len_b < 0) {
          av_log(NULL, AV_LOG_ERROR, "Audio convesion failed\n");
          break; // exit the loop
        }
        if (len_b == out_count) {
          swr_init(vs->swr_ctx);
          av_log(NULL, AV_LOG_WARNING, "Not enough Audio buffer\n");
        }

        // converted audio frame is stored in audio_buf
        vs->audio_buf = vs->audio_buf1;
        cv_data_size = len_b * vs->audio_tgt.channels * av_get_bytes_per_sample(vs->audio_tgt.fmt);
      } else {
        vs->audio_buf = vs->frame->data[0];
        cv_data_size = or_data_size;
      }

      //audio_clock0 = vs->audio_clock; 

      // update the audio clock with the pts 
      if (vs->frame->pts != AV_NOPTS_VALUE) // DEBUG (0, AV_NOPTS_VALUE)
        vs->audio_clock = vs->frame->pts * av_q2d(tb) + (double) vs->frame->nb_samples / vs->frame->sample_rate;
      else
        vs->audio_clock = NAN;

      vs->audio_clock_serial = vs->audio_pkt_tmp_serial;

      /* DEBUG -----------------------------------
      // show step size
         static double last_clock;
         printf("audio: delay=%0.3f clock=%0.3f clock0=%0.3f\n",
                vs->audio_clock - last_clock,
                vs->audio_clock, audio_clock0);
         last_clock = vs->audio_clock;
      // ---------------------------------------- */

      // return converted audio sample size and continue with the next frame of the audio packet
      return cv_data_size;
    }

    // free the current packet
    if(pkt->data)
      av_free_packet(pkt);
    // set the end to 0 to avoid overlapping
    memset(pkt_tmp, 0, sizeof(*pkt_tmp));

    if(vs->audioq.quit) 
      return -1;

    if (vs->audioq.nb_packets == 0)
      SDL_CondSignal(vs->continue_read_thread);

    // get next packet on quewe
    if ((new_pkt = packet_queue_get(&vs->audioq, pkt, 1, &vs->audio_pkt_tmp_serial)) < 0)
      return -1;

    if(pkt->data == flush_pkt.data) {
      avcodec_flush_buffers(dec);
      flush_done = 0;
      vs->audio_buf_frames_pending = 0;
    }

    *pkt_tmp = *pkt;  
  }
}

// prepare a new SDL audio buffer  (substitute  current_audio_buf)
static void sdl_audio_callback(void *userdata, Uint8 *stream, int len) {
  VideoState *vs = userdata;
  int64_t audio_callback_time;
  int audio_size, len_a;
  int bytes_per_sec, audio_buffer;
  double pts; 
  int frame_size = av_samples_get_buffer_size(NULL, 
                                              vs->audio_tgt.channels, 
                                              1, 
                                              vs->audio_tgt.fmt, 
                                              1);
  audio_callback_time = av_gettime();

  while(len > 0) {

    if(vs->audio_buf_index >= vs->audio_buf_size) {
       audio_size = audio_decode_frame(vs);
       if(audio_size < 0) {
	      // if error the output silence
	      vs->audio_buf      = vs->silence_buf;
	      vs->audio_buf_size = sizeof(vs->silence_buf) / frame_size * frame_size;
       } else {
	      vs->audio_buf_size = audio_size;
       }
       vs->audio_buf_index = 0;
    }

    len_a = vs->audio_buf_size - vs->audio_buf_index;

    if(len_a > len)
       len_a = len;

    memcpy(stream, (uint8_t *)vs->audio_buf + vs->audio_buf_index, len_a);

    len -= len_a;
    stream += len_a;
    vs->audio_buf_index += len_a;
  }
  // calculate the audio stream bytes per second
  bytes_per_sec = vs->audio_tgt.freq * vs->audio_tgt.channels * av_get_bytes_per_sample(vs->audio_tgt.fmt);
  audio_buffer = vs->audio_buf_size - vs->audio_buf_index;
  // Assume the audio driver that is used by SDL has two periods
  if (!isnan(vs->audio_clock)) {  
    //av_log(NULL, AV_LOG_WARNING, "valid value of audio_clock=%0.3f\n", vs->audio_clock);
    set_clock_at(&vs->audclk, vs->audio_clock - (double)(2 * vs->audio_hw_buf_size + audio_buffer) / bytes_per_sec, vs->audio_clock_serial, audio_callback_time / (double)AV_TIME_BASE);
    sync_clock_to_slave(&vs->extclk, &vs->audclk);
  }
}

static int audio_open(void *userdata, int64_t wanted_channel_layout, int wanted_nb_channels, int wanted_sample_rate, struct AudioParams *audio_hw_params) {

    SDL_AudioSpec sdl_audio, sdl_spec;
    const char *env;
    const int next_nb_channels[] = {0, 0, 1, 6, 2, 6, 4, 6};

    env = SDL_getenv("SDL_AUDIO_CHANNELS");
    if (env) {
        wanted_nb_channels = atoi(env);
        wanted_channel_layout = av_get_default_channel_layout(wanted_nb_channels);
    }
    if (!wanted_channel_layout || wanted_nb_channels != av_get_channel_layout_nb_channels(wanted_channel_layout)) {
        wanted_channel_layout = av_get_default_channel_layout(wanted_nb_channels);
        wanted_channel_layout &= ~AV_CH_LAYOUT_STEREO_DOWNMIX;
    }
    sdl_audio.channels = av_get_channel_layout_nb_channels(wanted_channel_layout);
    sdl_audio.freq = wanted_sample_rate;
    if (sdl_audio.freq <= 0 || sdl_audio.channels <= 0) {
        fprintf(stderr, "Invalid sample rate or channel count!\n");
        return -1;
    }
    sdl_audio.format   = AUDIO_S16SYS;
    sdl_audio.samples  = SDL_AUDIO_BUFFER_SIZE;
    sdl_audio.callback = sdl_audio_callback;
    sdl_audio.userdata = userdata;
    sdl_audio.silence  = 0;

    while (SDL_OpenAudio(&sdl_audio, &sdl_spec) < 0) {
        av_log(NULL, AV_LOG_WARNING, "SDL_OpenAudio (%d channels): %s\n", sdl_audio.channels, SDL_GetError());
        sdl_audio.channels = next_nb_channels[FFMIN(7, sdl_audio.channels)];
        if (!sdl_audio.channels) {
            av_log(NULL, AV_LOG_ERROR, "No more channel combinations to try, audio open failed\n");
            return -1;
        }
        wanted_channel_layout = av_get_default_channel_layout(sdl_audio.channels);
    }
    if (sdl_spec.format != AUDIO_S16SYS) {
         av_log(NULL, AV_LOG_ERROR, "SDL advised audio format %d is not supported!\n", sdl_spec.format);
        return -1;
    }
    if (sdl_spec.channels != sdl_audio.channels) {
        wanted_channel_layout = av_get_default_channel_layout(sdl_spec.channels);
        if (!wanted_channel_layout) {
            av_log(NULL, AV_LOG_ERROR, "SDL advised channel count %d is not supported!\n", sdl_spec.channels);
            return -1;
        }
    }

    audio_hw_params->fmt            = AV_SAMPLE_FMT_S16;
    audio_hw_params->freq           = sdl_spec.freq;
    audio_hw_params->channel_layout = wanted_channel_layout;
    audio_hw_params->channels       = sdl_spec.channels;
    return sdl_spec.size;
}


// open a given Audio or Video stream. Return 0 if opened without errors
static int stream_open(VideoState *vs, int stream_index) {

  AVFormatContext *avf_ctx = vs->avf_ctx;
  AVCodecContext *avc_ctx;
  AVCodec *codec;
  AVDictionary *opts; // NEW
  AVDictionaryEntry *t = NULL; // NEW
  int ret;
  int sample_rate, nb_channels;
  int64_t channel_layout;

  // restrict to a valid stream
  if(stream_index < 0 || stream_index >= avf_ctx->nb_streams)
     return -1;

  // get a pointer to the stream codec context
  avc_ctx = avf_ctx->streams[stream_index]->codec;
  // find the decoder for the given stream
  codec = avcodec_find_decoder(avc_ctx->codec_id);

  // overwrite codec id if specified by the user
  switch(avc_ctx->codec_type){
     case AVMEDIA_TYPE_AUDIO:
        if(audio_codec_name != NULL)
           if(codec == avcodec_find_decoder_by_name(audio_codec_name))
               av_log(NULL, AV_LOG_INFO, "Audio codec found, using requested %s audio codec\n", audio_codec_name);
           else
               av_log(NULL, AV_LOG_INFO, "Audio codec not found, using detected %s audio codec\n", codec->name);
        break;
     case AVMEDIA_TYPE_VIDEO:
        if(video_codec_name != NULL)
           if(codec == avcodec_find_decoder_by_name(video_codec_name))
              av_log(NULL, AV_LOG_INFO, "Video codec found, using requested %s video codec\n", video_codec_name);
           else
               av_log(NULL, AV_LOG_INFO, "Video codec not found, using detected %s video codec\n", codec->name);
        break;
  }
  if(!codec) {
     av_log(NULL, AV_LOG_WARNING, "Unable to find decoder!\n");
     return -1;
  }
  // set some fixed parameters  
  // FIXME with input data
  avc_ctx->codec_id          = codec->id;
  avc_ctx->workaround_bugs   = 1;
  avc_ctx->error_concealment = 3;
  avc_ctx->idct_algo         = FF_IDCT_AUTO;

  // set the initial value for lowres
  if(drop ==1){
     avc_ctx->lowres = drop;
     if(avc_ctx->lowres > codec->max_lowres) {
        av_log(avc_ctx, 
            AV_LOG_WARNING, 
            "The maximum LOW RESOURCES value supported by the decoder is %d\n",
             codec->max_lowres);
        // change the value of lowres
        avc_ctx->lowres = codec->max_lowres;
        avc_ctx->flags2 |= CODEC_FLAG2_FAST;
     }
  }
  //if(drop ==0)
  //   avc_ctx->flags2 |= CODEC_FLAG2_FAST;

  // Don't draw edges 
  if(avc_ctx->lowres)
     avc_ctx->flags |= CODEC_FLAG_EMU_EDGE;

  // NEVER draw edges (they will break buffer initialization)
  if(codec->capabilities & CODEC_CAP_DR1)
     avc_ctx->flags |= CODEC_FLAG_EMU_EDGE;
/*
  // Mpeg 1/2
  if(avc_ctx->codec_id==CODEC_ID_MPEG1VIDEO || avc_ctx->codec_id==CODEC_ID_MPEG2VIDEO)
     avc_ctx->flags |= CODEC_FLAG_INTERLACED_DCT;

  codec->capabilities &= ~CODEC_CAP_AUTO_THREADS;
  if(avc_ctx->thread_count == 0)
     codec->capabilities |= CODEC_CAP_AUTO_THREADS;
*/

  opts = filter_codec_opts(codec_opts, avc_ctx->codec_id, avf_ctx, avf_ctx->streams[stream_index], codec);

  if (!av_dict_get(opts, "threads", NULL, 0)) // NEW
     av_dict_set(&opts, "threads", "auto", 0);

  if (avc_ctx->lowres)
     av_dict_set(&opts, "lowres", av_asprintf("%d", avc_ctx->lowres), AV_DICT_DONT_STRDUP_VAL);

  if (avc_ctx->codec_type == AVMEDIA_TYPE_VIDEO || avc_ctx->codec_type == AVMEDIA_TYPE_AUDIO)
     av_dict_set(&opts, "refcounted_frames", "1", 0);

  // open the decoder for the given stream
  if (avcodec_open2(avc_ctx, codec, &opts) < 0) {
     av_log(NULL, AV_LOG_ERROR,"Unable to open codec!\n");
     return -1;
  }
  if ((t = av_dict_get(opts, "", NULL, AV_DICT_IGNORE_SUFFIX))) {
     av_log(NULL, AV_LOG_ERROR, "Option %s not found.\n", t->key);
     return AVERROR_OPTION_NOT_FOUND;
  }

  avf_ctx->streams[stream_index]->discard = AVDISCARD_DEFAULT;

  switch(avc_ctx->codec_type) {
  case AVMEDIA_TYPE_AUDIO:
#if HAVE_LIBAVFILTER
  {
     AVFilterLink *link;

     vs->audio_filter_src.freq           = avc_ctx->sample_rate;
     vs->audio_filter_src.channels       = avc_ctx->channels;
     vs->audio_filter_src.channel_layout = get_valid_channel_layout(avc_ctx->channel_layout, avc_ctx->channels);
     vs->audio_filter_src.fmt            = avc_ctx->sample_fmt;
     if ((ret = configure_audio_filters(vs, afilters, 0)) < 0)
       return ret;
     link = vs->out_audio_filter->inputs[0];
     sample_rate    = link->sample_rate;
     nb_channels    = link->channels;
     channel_layout = link->channel_layout;
  }
#else
     sample_rate    = avc_ctx->sample_rate;
     nb_channels    = avc_ctx->channels;
     channel_layout = avc_ctx->channel_layout;
#endif

    /* prepare audio output */
    if ((ret = audio_open(vs, channel_layout, nb_channels, sample_rate, &vs->audio_tgt)) < 0)
      return ret;

    vs->audio_hw_buf_size = ret;
    vs->audio_src = vs->audio_tgt;
    vs->audio_buf_size = 0;
    vs->audio_buf_index = 0;

    // averaging filter for audio syncronization
    vs->audio_diff_avg_coef = exp(log(0.01) / AUDIO_DIFF_AVG_NB);
    vs->audio_diff_avg_count = 0;
    // the audio correction is made only if larger error than this threshold
    // because there isn't enough precise audio fifo
    vs->audio_diff_threshold = 2.0 * vs->audio_hw_buf_size / av_samples_get_buffer_size(NULL, vs->audio_tgt.channels, vs->audio_tgt.freq, vs->audio_tgt.fmt, 1);

    memset(&vs->audio_pkt, 0, sizeof(vs->audio_pkt));
    memset(&vs->audio_pkt_tmp, 0, sizeof(vs->audio_pkt_tmp));

    vs->audio_stream = stream_index;
    vs->audio_st = avf_ctx->streams[stream_index];

    packet_queue_start(&vs->audioq);
    SDL_PauseAudio(0);
    break;
  case AVMEDIA_TYPE_VIDEO:
    vs->video_stream = stream_index;
    vs->video_st = avf_ctx->streams[stream_index];
    // initialize picture queue
    packet_queue_start(&vs->videoq);
    // call to create the SDL video thread
    vs->video_tid = SDL_CreateThread(video_thread, vs);
    vs->queue_attachments_req = 1;
    break;
  default:
    break;
  }
  return 0;
}

// safely close all streams
static void stream_close(VideoState *vs, int stream_index)
{
    AVFormatContext *avf_ctx = vs->avf_ctx;
    AVCodecContext *avc_ctx;

    if (stream_index < 0 || stream_index >= avf_ctx->nb_streams)
        return;
    avc_ctx = avf_ctx->streams[stream_index]->codec;

    switch(avc_ctx->codec_type) {
    case AVMEDIA_TYPE_AUDIO:
        packet_queue_quit(&vs->audioq);

        SDL_CloseAudio();

        packet_queue_flush(&vs->audioq);
        av_free_packet(&vs->audio_pkt);
        swr_free(&vs->swr_ctx);
        av_freep(&vs->audio_buf1);
        vs->audio_buf1_size = 0;
        vs->audio_buf = NULL;
        av_frame_free(&vs->frame);

        if (vs->rdft) {
            av_rdft_end(vs->rdft);
            av_freep(&vs->rdft_data);
            vs->rdft = NULL;
            vs->rdft_bits = 0;
        }
#if HAVE_LIBAVFILTER
        avfilter_graph_free(&vs->agraph);
#endif
        break;
    case AVMEDIA_TYPE_VIDEO:
        packet_queue_quit(&vs->videoq);
        // we also signal this mutex to make sure 
        // we deblock the video thread in all cases
        SDL_LockMutex(vs->pictq_mutex);
        SDL_CondSignal(vs->pictq_cond);
        SDL_UnlockMutex(vs->pictq_mutex);
        //Wait for the thread to finish
        SDL_WaitThread(vs->video_tid, NULL);

        packet_queue_flush(&vs->videoq);
        break;
    default:
        break;
    }

    avf_ctx->streams[stream_index]->discard = AVDISCARD_ALL;
    avcodec_close(avc_ctx);

    switch(avc_ctx->codec_type) {
    case AVMEDIA_TYPE_AUDIO:
        vs->audio_st = NULL;
        vs->audio_stream = -1;
        break;
    case AVMEDIA_TYPE_VIDEO:
        vs->video_st = NULL;
        vs->video_stream = -1;
        break;
    default:
        break;
    }

}

void print_error(const char *filename, int err) {

    char errbuf[128];
    const char *errbuf_ptr = errbuf;

    if (av_strerror(err, errbuf, sizeof(errbuf)) < 0)
        errbuf_ptr = strerror(AVUNERROR(err));
    av_log(NULL, AV_LOG_ERROR, "%s: %s\n", filename, errbuf_ptr);
}

static int decode_interrupt_cb(void *ctx) {

  VideoState *vs = ctx;
  return vs->quit;
}

// real time processing
static int is_realtime(AVFormatContext *s) {

    if(   !strcmp(s->iformat->name, "rtp")
       || !strcmp(s->iformat->name, "rtsp")
       || !strcmp(s->iformat->name, "sdp")  )
        return 1;

    if(s->pb && (   !strncmp(s->filename, "rtp:", 4)
                 || !strncmp(s->filename, "udp:", 4)  )  )
        return 1;

    return 0;
}

// this thread gets the stream passed by SDL
static int decode_thread(void *arg) {

  VideoState *vs = arg;
  AVFormatContext *avf_ctx = NULL;
  AVCodecContext *avctx;
  AVCodec *codec;
  AVPacket pkt1, *packet = &pkt1;
  SDL_mutex *wait_mutex = SDL_CreateMutex();

  int64_t stream_time_s;
  int video_index = -1;
  int audio_index = -1;
  int pkt_user_range = 0;
  int flag_broken_video = 0;
  int i, ret, err;
  int eof = 0;
  AVDictionaryEntry *t; // NEW
  AVDictionary **opts; // NEW
  int orig_nb_streams; // NEW

  vs->last_video_stream = vs->video_stream = -1;
  vs->last_audio_stream = vs->audio_stream = -1;

  avf_ctx = avformat_alloc_context();
  // This will interrupt blocking functions when we quit!
  avf_ctx->interrupt_callback.callback = decode_interrupt_cb;
  avf_ctx->interrupt_callback.opaque = vs;

  // Open video file
  err = avformat_open_input(&avf_ctx, vs->filename, vs->iformat, &format_opts);
  if (err < 0) {
     print_error(vs->filename, err);
     ret = -1;
     goto fail; // Couldn't open file
  }
  if ((t = av_dict_get(format_opts, "", NULL, AV_DICT_IGNORE_SUFFIX))) { // NEW
     av_log(NULL, AV_LOG_ERROR, "Option %s not found.\n", t->key);
     ret = AVERROR_OPTION_NOT_FOUND;
     goto fail;
  }

  // register VideoState AVformatContext 
  vs->avf_ctx = avf_ctx;

  opts = setup_find_stream_info_opts(avf_ctx, codec_opts);  // NEW
  orig_nb_streams = avf_ctx->nb_streams;

  // Retrieve stream information
  err = avformat_find_stream_info(avf_ctx, opts); // changed NULL -> opts
  if (err < 0) {
     fprintf(stderr, "%s: could not find stream info\n", vs->filename);
     ret = -1;
     goto fail; // Couldn't find stream information
  }
  for (i = 0; i < orig_nb_streams; i++)  // NEW
     av_dict_free(&opts[i]);
  av_freep(&opts);


  // FIXME the player should not use url_feof() to test for the end
  if(avf_ctx->pb) 
     avf_ctx->pb->eof_reached = 0;

  // set the maximum frame duration that will be used to refresh the video rate
  vs->max_frame_duration = (avf_ctx->iformat->flags & AVFMT_TS_DISCONT) ? 10.0 : 3600.0;

  // if start time value is given
  if (time_s != AV_NOPTS_VALUE) {
     int64_t time_initial;

     time_initial = time_s;
     // Add the stream start-time
     if (avf_ctx->start_time != AV_NOPTS_VALUE)
        time_initial += avf_ctx->start_time;

     ret = avformat_seek_file(avf_ctx, -1, INT64_MIN, time_initial, INT64_MAX, 0);
     if (ret < 0) {
        fprintf(stderr, "%s: Unable to start from position %0.3f\n",
                vs->filename, (double)time_initial / AV_TIME_BASE);
     }
  }

  // find which realtime code is to be used
  vs->realtime = is_realtime(avf_ctx);

  for(i=0; i<avf_ctx->nb_streams; i++)
     avf_ctx->streams[i]->discard = AVDISCARD_ALL;

  // Find best video stream
  if(disable_video==0) {
     video_index = av_find_best_stream(avf_ctx, AVMEDIA_TYPE_VIDEO,
                                       -1, 
                                       -1, 
                                       NULL, 0);
  }
  // Find best audio stream
  if(disable_audio==0) {
     audio_index = av_find_best_stream(avf_ctx, AVMEDIA_TYPE_AUDIO,
                                       -1,
                                       video_index,
                                       NULL, 0);
  }

/*
  for(i=0; i<avf_ctx->nb_streams; i++) {
    avf_ctx->streams[i]->discard = AVDISCARD_ALL;

    avctx = avf_ctx->streams[i]->codec;
  // Find the first audio stream
      if(avctx->codec_type == AVMEDIA_TYPE_AUDIO && audio_index < 0) {
        audio_index=i;
        fprintf(stderr, "found audio stream at index: %i\n", audio_index);
      }
  // Find the first video stream
      if(avctx->codec_type == AVMEDIA_TYPE_VIDEO && video_index < 0) {
        video_index=i;
        fprintf(stderr, "found video stream at index: %i\n", video_index);
      }
  }
*/

  // Open the audio stream
  if(audio_index >= 0 && disable_audio == 0) {
     av_log(NULL, AV_LOG_INFO, "found Audio stream at index %i\n", audio_index);
     stream_open(vs, audio_index);
  }
  // Open the video stream
  if(video_index >= 0 && disable_video == 0) {
     av_log(NULL, AV_LOG_INFO, "found Video stream at index %i\n", video_index);
     stream_open(vs, video_index);
  }  

  if(vs->video_stream < 0 && vs->audio_stream < 0) {
     av_log(NULL, AV_LOG_FATAL, "%s: Unable to find any stream\n", vs->filename);
     ret = -1;
     goto fail;
  }

  // do not limit buffer filling in case realtime option is selected: dangerous
  if (nolimit_buffer == 0 && vs->realtime)
     nolimit_buffer = 1;

  // The main decoding loop
  for(;;) {
     if(vs->quit)
        break;

     if (vs->pause_curr != vs->pause_last) {
        vs->pause_last = vs->pause_curr;

        if(vs->pause_curr)
           vs->pause_back = av_read_pause(avf_ctx);
        else
           av_read_play(avf_ctx);
     } 

     // wait 10 ms to avoid trying to get another packet
     if (vs->pause_curr && 
             (!strcmp(avf_ctx->iformat->name, "rtsp") ||
             (avf_ctx->pb ))) {
         SDL_Delay(10);
         continue;
     }
     ret = -1;
     // the seeking subroutine 
     if(vs->seek_req) {
        int64_t seek_target = vs->seek_pos;
        int64_t seek_min    = vs->seek_rel > 0 ? seek_target - vs->seek_rel + 2: INT64_MIN;
        int64_t seek_max    = vs->seek_rel < 0 ? seek_target - vs->seek_rel - 2: INT64_MAX;
        // FIXME the +-2 is due to rounding being not done in the correct direction in generation
        // of the seek_pos/seek_rel variables
        ret = avformat_seek_file(vs->avf_ctx, -1, seek_min, seek_target, seek_max, 0);
        if(ret < 0) {
	       av_log(NULL, AV_LOG_WARNING, "error while seeking on %s: \n", vs->avf_ctx->filename);
        } else {
		   // send audio packets to queue
	       if(disable_audio == 0 && vs->audio_stream >= 0) {
	          packet_queue_flush(&vs->audioq);
	          packet_queue_put(&vs->audioq, &flush_pkt);
              av_log(NULL, AV_LOG_INFO, "Seeking on audio stream: %i\n", vs->audio_stream);
	       }
		   // send video packets to queue
           if(disable_video == 0 && vs->video_stream >= 0) {
	          packet_queue_flush(&vs->videoq);
	          packet_queue_put(&vs->videoq, &flush_pkt);
              av_log(NULL, AV_LOG_INFO, "Seeking on video stream: %i\n", vs->video_stream);
	       } 
           set_clock(&vs->extclk, seek_target / (double)AV_TIME_BASE, 0);
        }
        // reset seek count and prepare for next request
        vs->seek_req = 0;
        vs->queue_attachments_req = 1;
        eof= 0;
        if (vs->pause_curr)
           step_to_next_frame(vs);
    }

    if (vs->queue_attachments_req) {
        //av_log(NULL, AV_LOG_INFO, "Requested copy of video packet\n");
        if (vs->video_st && vs->video_st->disposition & AV_DISPOSITION_ATTACHED_PIC) { //FIXME
            AVPacket copy;
            if ((ret = av_copy_packet(&copy, &vs->video_st->attached_pic)) < 0){
                av_log(NULL, AV_LOG_WARNING, "Copy video packet failed\n");
                goto fail;
            }
            packet_queue_put(&vs->videoq, &copy);
            av_log(NULL, AV_LOG_INFO, "Copy video packet success\n");
        }
        vs->queue_attachments_req = 0;
    }

    // if the queue are full, no need to read more
    if(nolimit_buffer==0 
       && (vs->audioq.size + vs->videoq.size > MAX_QUEUE_SIZE 
           || (  (vs->audioq.nb_packets > MIN_FRAMES || vs->audio_stream < 0 || vs->audioq.quit)
               &&(vs->videoq.nb_packets > MIN_FRAMES || vs->video_stream < 0 || vs->videoq.quit
                  || (vs->video_st->disposition & AV_DISPOSITION_ATTACHED_PIC))))){  // FIXME
       // wait 10 msec
       SDL_LockMutex(wait_mutex);
       SDL_CondWaitTimeout(vs->continue_read_thread, wait_mutex, 10);
       SDL_UnlockMutex(wait_mutex);
       continue;
    }

    // check if we are at the end and reset to initial values
    if (eof) {
       av_log(NULL, AV_LOG_INFO, "EOF: reached\n");
       if(vs->video_stream >= 0){
           av_init_packet(packet);
           packet->data = NULL;
           packet->size = 0;
           packet->stream_index = vs->video_stream;
           packet_queue_put(&vs->videoq, packet);
        } 
        if(vs->audio_stream >= 0){
           av_init_packet(packet);
           packet->data = NULL;
           packet->size = 0;
           packet->stream_index = vs->audio_stream;
           packet_queue_put(&vs->audioq, packet);
        }
        SDL_Delay(10);
        if (vs->audioq.size + vs->videoq.size == 0) {
           if(loop == 1 ){
              av_log(NULL, AV_LOG_INFO, "EOF: loop begin at time=%d\n", time_s);
              stream_seek(vs, time_s != AV_NOPTS_VALUE ? time_s : 0, 0);  //FIXME
           } else if(loop == 0 ){
              ret=AVERROR_EOF;
              goto fail;
           }
        } 
        eof=0;
        continue;
    }

    ret = av_read_frame(avf_ctx, packet);
    if(ret < 0) {
        if(url_feof(avf_ctx->pb) || ret == AVERROR_EOF){
            eof=1;
            av_log(NULL, AV_LOG_WARNING, "AVERROR_EOF\n");
        }
        if(avf_ctx->pb && avf_ctx->pb->error) {
            av_log(NULL, AV_LOG_ERROR, "AVFormatContext->pb->error\n");
            break;
        }
        SDL_LockMutex(wait_mutex);
        SDL_CondWaitTimeout(vs->continue_read_thread, wait_mutex, 10);
        SDL_UnlockMutex(wait_mutex);
        continue;
    }

    // check if packet within the DURATION specified by user, then queue, otherwise discard 
    stream_time_s = avf_ctx->streams[packet->stream_index]->start_time;
    pkt_user_range = time_d == AV_NOPTS_VALUE ||  
                   (packet->pts - (stream_time_s != AV_NOPTS_VALUE ? stream_time_s : 0)) *
                   av_q2d(avf_ctx->streams[packet->stream_index]->time_base) -
                   (double)(time_s != AV_NOPTS_VALUE ? time_s : 0) / (double)AV_TIME_BASE
                   <= ((double)time_d / (double)AV_TIME_BASE);

    // get a pointer to the stream codec context
    avctx = avf_ctx->streams[packet->stream_index]->codec;
    // find the decoder for the given stream
    codec = avcodec_find_decoder(avctx->codec_id);
    // fix behaviour for some old video codec

    if(avctx->codec_type == AVMEDIA_TYPE_VIDEO) {
       if(codec == avcodec_find_decoder_by_name("mpeg1video")
       || codec == avcodec_find_decoder_by_name("mpeg2video")
       || codec == avcodec_find_decoder_by_name("wmv3")) {
           pkt_user_range = time_d == AV_NOPTS_VALUE || time_d; // BROKEN VIDEO HEADER FIX: setting time to start
           flag_broken_video = 1;
       }
    }

    // check if packet is from the audio or video stream
    if(packet->stream_index == vs->audio_stream && pkt_user_range) {
       packet_queue_put(&vs->audioq, packet);
       // av_log(NULL, AV_LOG_INFO, "is video pkt in user range?%i\n", pkt_user_range);
    } else 
    if(packet->stream_index == vs->video_stream && pkt_user_range
       && !(vs->video_st->disposition & AV_DISPOSITION_ATTACHED_PIC)) { // FIXME
       packet_queue_put(&vs->videoq, packet);
       //av_log(NULL, AV_LOG_INFO, "is audio pkt in user range?=%i\n", pkt_user_range);
    } else {
       av_free_packet(packet);
       av_log(NULL, AV_LOG_INFO, "no packet in the user range -> drop packet\n");

       // if at the end of user range, wait 10 msec and then play in loop if requested
       SDL_Delay(10);
       if(flag_broken_video == 1){   // BROKEN VIDEO HEADER FIX: setting time to start after a valid time duration
           printf("vs->audioq.size=%d\n", vs->audioq.size);
           printf("vs->videoq.size=%d\n", vs->videoq.size);
           vs->audioq.size = vs->videoq.size = 0;
           if(loop == 1 ){
               printf("broken video: loop begin at time=%d\n", time_s);
               stream_seek(vs, time_s != 0 ? time_s : 0, 0);
           }
       } else { 
          if(vs->audioq.size + vs->videoq.size == 0){ 
             if(loop == 1 ){
                printf("loop begin at time=%d\n", time_s);
                stream_seek(vs, time_s != 0 ? time_s : 0, 0);
             }
          }
       }
       continue;
    }
  }

  // wait until the end
  while(!vs->quit) {
     SDL_Delay(100);
  }
  ret = 0;

 fail:
    printf("close audio and video streams\n");
    if (vs->audio_stream >= 0)
       stream_close(vs, vs->audio_stream);      
    if (vs->video_stream >= 0)
       stream_close(vs, vs->video_stream);
    if (vs->avf_ctx)
       avformat_close_input(&vs->avf_ctx);

    if(ret != 0) {
       SDL_Event event;
       event.type = FF_QUIT_EVENT;
       event.user.data1 = vs;
       SDL_PushEvent(&event);
    }
    SDL_DestroyMutex(wait_mutex);
    return 0;
}

static void do_video_frame_drop(VideoState *vs) {
    double remaining_time = 0.00;  // DEBUG
    if (drop==0) {
      drop=1;
      fprintf(stderr, "video frame drop: enabled\n");
	} else if (drop==1) {
      drop=0;
      fprintf(stderr, "video frame drop: disabled\n");
    }
    video_refresh_timer(vs, &remaining_time);
}

static void do_full_screen(VideoState *vs) {
    if (full==0) 
      full=1;
	else if (full==1)
      full=0;
#if defined HAVE_APPLE 
    int i;
    for (i = 0; i < VIDEO_PICTURE_QUEUE_SIZE; i++) {
        vs->pictq[i].reallocate = 1;
    }
#endif
    video_screen_open(vs, 1, NULL);
}

static void do_cycle_audio_ch(VideoState *vs) {
    int str_index;
    int nb_channels;
    AVFormatContext *avf_ctx = vs->avf_ctx;
    AVCodecContext *dec = vs->audio_st->codec;
    AVStream *st = vs->audio_st;
 
    if (disable_audio==0 && st->codec->sample_rate != 0) {
      str_index=vs->audio_stream;
      nb_channels=dec->channels;
      fprintf(stderr, "total audio channels: %i\n", nb_channels);

      if(cycle_audio_ch >=0 && cycle_audio_ch < nb_channels)
         cycle_audio_ch ++;
      else
         cycle_audio_ch=0;

      dec->channel_layout = cycle_audio_ch;
      av_get_channel_layout_nb_channels(dec->channel_layout);
      fprintf(stderr, "requested audio channel: %i\n", av_get_channel_layout_nb_channels(dec->channel_layout));

      stream_close(vs, str_index);
      stream_open(vs, str_index);
    }
}

// useful if used together with audio delay command
static void do_change_refresh(VideoState *vs, int tune) {
    switch(tune){
      case -1:
        tunerefresh -= 1e5;
        if(tunerefresh < -(double)AV_TIME_BASE)
          tunerefresh = -(double)AV_TIME_BASE;
        tunestep -= 1;
        if(tunestep < -10)
          tunestep = -10;
        break;
      case 0:
        break;
      case 1:
        tunerefresh += 1e5;
        if(tunerefresh > (double)AV_TIME_BASE)
          tunerefresh = (double)AV_TIME_BASE;
        tunestep += 1;
        if(tunestep > 10)
          tunestep = 10;
        break;
      default:
        break;
    }
    video_command_icon(tunestep);
    //fprintf(stderr, "imgcommand(return): %s, imgtime(return): %s\n", imgcommand, imgtime);
    show_command = 1;
    SDL_CondSignal(vs->continue_read_thread);
}


static void refresh_loop_wait_event(VideoState *vs, SDL_Event *event) {
    double remaining_time = 0.0;
    SDL_PumpEvents();
    while (!SDL_PeepEvents(event, 1, SDL_GETEVENT, SDL_ALLEVENTS)) {
        if (!cursor_hidden && av_gettime() - cursor_last_shown > CURSOR_HIDE_DELAY) {
            SDL_ShowCursor(0);
            cursor_hidden = 1;
        }
        if (remaining_time > 0.0)
            av_usleep((int64_t)(remaining_time * ((double)AV_TIME_BASE + tunerefresh)));
        //fprintf(stderr, "tunerefresh: %f\n", tunerefresh);
        remaining_time = REFRESH_RATE;
        if (disable_video==0  && (!vs->pause_curr || vs->force_refresh))
            video_refresh_timer(vs, &remaining_time);
        SDL_PumpEvents();
    }
}

/* loop to handle an event sent by the GUI */
static int event_loop(VideoState *vs, int bpp) {
    SDL_Event event;
    double incr, pos;
    //int sflag = 0;
    for(;;) {
      double x;
      refresh_loop_wait_event(vs, &event);
      switch(event.type) {
      case SDL_KEYDOWN:
       switch(event.key.keysym.sym) {
         case SDLK_ESCAPE:
         case SDLK_q: // q: quit SDL loop
           do_exit(vs);
           return 1;
           break;
         case SDLK_SPACE:
         case SDLK_p: //p: pause playing
           do_pause(vs);
           break;
         case SDLK_s: //s: step to next frame
           step_to_next_frame(vs);
           break;
         case SDLK_a: //a: cicle audio channels
           do_cycle_audio_ch(vs);
           break;
         case SDLK_f: //f: fullscreen size
           do_full_screen(vs);
           vs->force_refresh = 1;
           break;
         case SDLK_d: //d: video frame drop
           do_video_frame_drop(vs);
           break;
         case SDLK_h: //h: manually increase horizontal size of the picture in case of wrong aspect ratio
           if (disable_video == 0){
             enlarge_w++;
             if(enlarge_w>9)
               enlarge_w=0;
             VideoPicture *vp;
             vp = &vs->pictq[vs->pictq_windex];
             video_screen_open(vs, 0, vp);
           }
           break;
         case SDLK_j: //j: manually decrease horizontal size of the picture in case of wrong aspect ratio
           if (disable_video == 0){
             enlarge_w--;
             if(enlarge_w<-9)
               enlarge_w=0;
             VideoPicture *vp;
             vp = &vs->pictq[vs->pictq_windex];
             video_screen_open(vs, 0, vp);
           }
           break;
         case SDLK_v: //v: manually increase vertical size of the picture in case of wrong aspect ratio
           if (disable_video == 0){
             enlarge_h++;
             if(enlarge_h>9)
               enlarge_h=0;
             VideoPicture *vp;
             vp = &vs->pictq[vs->pictq_windex];
             video_screen_open(vs, 0, vp);
           }
           break;
         case SDLK_c: //v: manually decrease vertical size of the picture in case of wrong aspect ratio
           if (disable_video == 0){
             enlarge_h--;
             if(enlarge_h<-9)
               enlarge_h=0;
             VideoPicture *vp;
             vp = &vs->pictq[vs->pictq_windex];
             video_screen_open(vs, 0, vp);
           }
           break;
         case SDLK_o: // o: original picture size
           if (disable_video == 0){
             screen_w = screen_h = 0;
             enlarge_w = enlarge_h = 0;
             imgcommand = ICON_SO;
             show_command = 1;
             VideoPicture *vp;
             vp = &vs->pictq[vs->pictq_windex];
             video_screen_open(vs, 0, vp);
           }
           break;
         case SDLK_b: // b: bigger picture size
           if (disable_video == 0){
/*
               if(vs->video_st->codec->width && !screen_w) 
                 screen_w = vs->video_st->codec->width;
               if(vs->video_st->codec->height && !screen_h)
                 screen_h = vs->video_st->codec->height;

               screen_w = vs->width = screen_w * 1.1;
               screen_h = vs->height = screen_h * 1.1;
               screen = SDL_SetVideoMode (screen_w, screen_h, bpp, SDL_HWSURFACE|SDL_ASYNCBLIT|SDL_HWACCEL);

               if(screen_w > vs->video_st->codec->width) 
                 imgcommand = ICON_SG;
               else if(screen_w < vs->video_st->codec->width) 
                 imgcommand = ICON_SS;
               else if(screen_w == vs->video_st->codec->width) 
                 imgcommand = ICON_SO;
               show_command = 1;
*/
             enlarge_w++;
             enlarge_h++;
             if(enlarge_h>9 || enlarge_w>9) {
               enlarge_w=0;
               enlarge_h=0;
             }
             if(enlarge_w < 0) 
               imgcommand = ICON_SS;
             else if(enlarge_w > 0) 
               imgcommand = ICON_SG;
             else if(enlarge_w == 0) 
               imgcommand = ICON_SO;
             show_command = 1;
             VideoPicture *vp;
             vp = &vs->pictq[vs->pictq_windex];
             video_screen_open(vs, 0, vp);
           }
           break;
         case SDLK_g: // g: smaller picture size
           if (disable_video == 0) {
/*
                 if(vs->video_st->codec->width && !screen_w) 
                   screen_w = vs->video_st->codec->width;
                 if(vs->video_st->codec->height && !screen_h)
                   screen_h = vs->video_st->codec->height;

                 screen_w = vs->width  = screen_w / 1.1;
                 screen_h = vs->height = screen_h / 1.1;
                 screen = SDL_SetVideoMode (screen_w, screen_h, bpp, SDL_HWSURFACE|SDL_ASYNCBLIT|SDL_HWACCEL);

                 if(screen_w < vs->video_st->codec->width) 
                   imgcommand = ICON_SS;
                 else if(screen_w > vs->video_st->codec->width) 
                   imgcommand = ICON_SG;
                 else if(screen_w == vs->video_st->codec->width) 
                   imgcommand = ICON_SO;
                 show_command = 1;
*/
             enlarge_w--;
             enlarge_h--;
             if(enlarge_h<-9 || enlarge_w<-9) {
               enlarge_w=0;
               enlarge_h=0;
             }
             if(enlarge_w < 0) 
               imgcommand = ICON_SS;
             else if(enlarge_w > 0) 
               imgcommand = ICON_SG;
             else if(enlarge_w == 0) 
               imgcommand = ICON_SO;
             show_command = 1;
             VideoPicture *vp;
             vp = &vs->pictq[vs->pictq_windex];
             video_screen_open(vs, 0, vp);

           }
           break;
         case SDLK_0:
#ifdef ENABLE_AMIXER
        alsastep+=5;
        if(alsastep>100)
           alsastep=100;
        alsa_volume_control(alsastep);
	show_command=1;
        system(alsa_volume);
        //fprintf(stderr, "alsa volume control: %s\n", alsa_volume);
#endif
           break;
         case SDLK_9:
#ifdef ENABLE_AMIXER
        alsastep-=5;
        if(alsastep<0)
           alsastep=0;
        alsa_volume_control(alsastep);
	show_command=1;
        system(alsa_volume);
        //fprintf(stderr, "alsa volume control: %s\n", alsa_volume);
#endif
           break;
         case SDLK_7:
           do_change_refresh(vs, 1);
           break;
         case SDLK_8:
           do_change_refresh(vs, -1);
           break;
         case SDLK_RIGHT:
           incr = 10.0;
           imgcommand = ICON_F;
           show_command = 1;
           goto do_seek;
         case SDLK_LEFT:
           incr = -10.0;
           imgcommand = ICON_B;
           show_command = 1;
           goto do_seek;
         case SDLK_UP:
           incr = 60.0;
           imgcommand = ICON_FF;
           show_command = 1;
           goto do_seek;
         case SDLK_DOWN:
           incr = -60.0;
           imgcommand = ICON_BB;
           show_command = 1;
           goto do_seek;
         case SDLK_PAGEUP:
           incr = 600.0;
           imgcommand = ICON_FF;
           show_command = 1;
           goto do_seek;
         case SDLK_PAGEDOWN:
           incr = -600.0;
           imgcommand = ICON_BB;
           show_command = 1;
           goto do_seek;
         do_seek:
           if(vs){
             pos = get_master_clock(vs);
             //av_log(NULL, AV_LOG_INFO, "try: get master_clock pos=%0.3f\n", vs->audclk, pos);
             if(isnan(pos)){
                pos = (double)vs->seek_pos / (double)AV_TIME_BASE;
                //av_log(NULL, AV_LOG_INFO, "failed: pos=%0.3f, vs->seek_pos=%0.3f\n", pos, vs->seek_pos);
             }
             pos += incr;
             //av_log(NULL, AV_LOG_INFO, "incr=%0.3f, pos=%0.3f\n", incr, pos);
             if(pos < vs->avf_ctx->start_time / (double)AV_TIME_BASE && vs->avf_ctx->start_time != AV_NOPTS_VALUE){ 
                pos = vs->avf_ctx->start_time / (double)AV_TIME_BASE;
                av_log(NULL, AV_LOG_INFO, "failed: get_start_time, pos=%0.3f\n", pos);
             }
             stream_seek(vs, (int64_t)(pos * AV_TIME_BASE), (int64_t)(incr * AV_TIME_BASE));
             av_log(NULL, AV_LOG_INFO, "Seeking incr=%0.3f pos=%0.3f\n", incr * AV_TIME_BASE, pos * AV_TIME_BASE);
           }
           break;
         default:
           break;
      }
      break;
    case SDL_VIDEOEXPOSE:
      vs->force_refresh = 1;
      break; 
    case SDL_VIDEORESIZE:
      screen = SDL_SetVideoMode(FFMIN(16383, event.resize.w), event.resize.h, bpp,
                                SDL_HWSURFACE|SDL_ASYNCBLIT|SDL_RESIZABLE|SDL_HWACCEL);
      screen_w = vs->width = event.resize.w;
      screen_h = vs->height = event.resize.h;
      vs->force_refresh = 1;
      break;
    case SDL_QUIT:
    case FF_QUIT_EVENT:
      do_exit(vs);
      return 1;
      break;
    case FF_ALLOC_EVENT:
      alloc_picture(event.user.data1);
      break;
    default:
      break;
    }
  }

}

static int lockmgr(void **mtx, enum AVLockOp op) {
   switch(op) {
      case AV_LOCK_CREATE:
          *mtx = SDL_CreateMutex();
          if(!*mtx)
              return 1;
          return 0;
      case AV_LOCK_OBTAIN:
          return !!SDL_LockMutex(*mtx);
      case AV_LOCK_RELEASE:
          return !!SDL_UnlockMutex(*mtx);
      case AV_LOCK_DESTROY:
          SDL_DestroyMutex(*mtx);
          return 0;
   }
   return 1;
}

// parse file and call the GUI loop 
//                (filename,   starttime,  duration,  width, height, loop,  full,  drop,  mute, blind, clock, audiodelay, audiocodec, videocodec, filters)
int loadPlayerFile(char *name, int64_t ts, int64_t td, int w, int h, int l, int f, int d, int a, int v, int c, float ta, int as, char *acn, char *vcn, char *filter){
  printf("[1] start=%d, duration=%d, width=%i, height=%i, loop=%i, full=%i, drop=%i, mute=%i, blind=%i, clock=%i, audelay=%f, volume=%i, audio=%s, video=%s, filter=%s\n", ts, td, w, h, l, f, d, a, v, c, ta, as, acn, vcn, filter);
  Uint32 sdl_init_flags;
  Uint32 sdl_video_flags;
  const SDL_VideoInfo *sdl_vi;
  int sdl_bpp;
  VideoState  *vs;
  int sw, sh;
  cursor_hidden = 0;
  imgcommand = NULL;
  imgtime = NULL;
  audio_codec_name = NULL;
  video_codec_name = NULL;
  vfilters = NULL;
  afilters = NULL;
  time_s = time_d = 0;
  // starttime, duration, audiodelay
  if(ts > 0) time_s = (double)AV_TIME_BASE * ts; // 1e6
  if(td > 0) time_d = (double)AV_TIME_BASE * td; // 1e6
  if(ta != 0) time_a = ta / 200;
  // loop, fullscreen, framedrop
  if(l == 0 || l == 1) loop = l;
  if(f == 0 || f == 1) full = f;
  if(d == 0 || d == 1) drop = d;
  // disable audio or video
  if(a == 0 || a == 1) disable_audio = a;
  if(v == 0 || v == 1) disable_video = v;
  // codec requested by the user
  if(acn) 
    audio_codec_name = acn;
  if(vcn) 
    video_codec_name = vcn;
  if(filter) 
    vfilters = filter;
  if(0>= as <= 100)
    alsastep=as;

  av_log_set_flags(AV_LOG_SKIP_REPEATED);

  // Register codecs, filters, devices
#if HAVE_LIBAVCODEC
  avcodec_register_all();
#endif
#if HAVE_LIBAVDEVICE
  avdevice_register_all();
#endif
#if HAVE_LIBAVFILTER
  avfilter_register_all();
#endif
  // Register all for safety 
  av_register_all();
  // Initialize the network protocol
  avformat_network_init();

  init_opts();

  // set SDL init flags
  sdl_init_flags = SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_TIMER | SDL_INIT_EVENTTHREAD; // always initialize video
  if(disable_audio==1)
      sdl_init_flags &= ~SDL_INIT_AUDIO;
#if defined HAVE_MINGW32 || defined HAVE_APPLE
  sdl_init_flags &= ~SDL_INIT_EVENTTHREAD; // Not supported on Windows or Mac OS X
#endif

  if(SDL_Init(sdl_init_flags)) {
      av_log(NULL, AV_LOG_FATAL, "SDL: Unable to initialize: %s\n", SDL_GetError());
      return 1;
  }

  // an ABS VALUE video size is passed by user
  if(w>=480 && h>=360) {
      sw = screen_w = w; // printf("absolute width=%i\n", sw);
      sh = screen_h = h; // printf("absolute height=%i\n", sh);
  } else
  // an INCREMENT video size is passed by user
  if(-3 <= w <= 8) {
      increment_step = w;
      sw=0; // printf("safety width=%i\n", sw);
      sh=0; // printf("safety height=%i\n", sh);
      //fprintf(stderr, "incremented size by a factor of %i\n", w);
  } else
  // if none RECALL previously STORED video size
  if(screen_w>64 && screen_h>64) {
      sw = screen_w;
      sh = screen_h;
  }

#if HAVE_STRUCT_SDL_VIDEOINFOP
   sdl_vi = SDL_GetVideoInfo();
  // SDL find the display size
  if (disable_video == 0) {
      fs_screen_w = sdl_vi->current_w;
      fs_screen_h = sdl_vi->current_h;
  }
  // SDL find the bpp mode
  switch (sdl_vi->vfmt->BitsPerPixel) {
     case 16:
     case 24:
     case 32:
        sdl_bpp = sdl_vi->vfmt->BitsPerPixel;
        break;
     default:
        sdl_bpp = 0;
        break;
  }
  av_log(NULL, AV_LOG_INFO, "SDL bpp mode: %i\n", sdl_bpp);
  av_log(NULL, AV_LOG_INFO, "SDL full screen: w=%i, h=%i\n", fs_screen_w, fs_screen_h);

  if (fs_screen_w && full==1) {
    sw = fs_screen_w;
    sh = fs_screen_h;
  }
#else
  sdl_bpp = 0;
#endif

  // if requested start in full screen mode
  if (full==0)
    sdl_video_flags = SDL_HWSURFACE|SDL_ASYNCBLIT|SDL_RESIZABLE|SDL_HWACCEL;
  else
    sdl_video_flags = SDL_HWSURFACE|SDL_ASYNCBLIT|SDL_FULLSCREEN|SDL_HWACCEL;

  // despite VIDEO is enabled or not, we ALWAYS need to set SDL video mode to safely exit the queue loop.
  screen = SDL_SetVideoMode(sw, sh, sdl_bpp, sdl_video_flags);
  // if not set reset size
  if (screen == NULL){
    av_log(NULL, AV_LOG_ERROR, "Unable to set w=%i, h=%i, bpp=%i SDL video mode: %s\n", sw, sh, sdl_bpp, SDL_GetError());
    sw = sh = 0;
    //do_exit(vs);
    return 1;
  }

  SDL_EventState(SDL_ACTIVEEVENT, SDL_IGNORE);
  SDL_EventState(SDL_SYSWMEVENT, SDL_IGNORE);
  SDL_EventState(SDL_USEREVENT, SDL_IGNORE);

  if (av_lockmgr_register(lockmgr)) {
     av_log(NULL, AV_LOG_FATAL, "Unable to initialize lock manager!\n");
     do_exit(NULL);
  }

  av_init_packet(&flush_pkt);
  flush_pkt.data = (uint8_t *)&flush_pkt;


  // allocate filename memory
  if (vs = av_mallocz(sizeof(VideoState))) {
      stpncpy(vs->filename, name, sizeof(vs->filename));
  }
  else
      do_exit(NULL);
  // check for global structure
  if (!vs) {
      av_log(NULL, AV_LOG_FATAL, "VideoState: Unable to initialize struct!\n");
      do_exit(NULL);
  }
  // synchronize with audio clock first, then try video clock
  if (c==0 && disable_audio==0){
     av_sync_type = AV_SYNC_AUDIO_MASTER;
     av_log(NULL, AV_LOG_INFO, "Synchronized with AUDIO clock\n");
  }else if (c==1 && disable_video==0){
     av_sync_type = AV_SYNC_VIDEO_MASTER;
     av_log(NULL, AV_LOG_INFO, "Synchronized with VIDEO clock\n");
  }else if(c==2){
     av_sync_type = AV_SYNC_EXTERNAL_MASTER;
     av_log(NULL, AV_LOG_INFO, "Synchronized with EXTERNAK clock\n");
  }

  // syncronize
  vs->av_sync_type = av_sync_type;
  vs->iformat = iformat;
  vs->ytop    = 0;
  vs->xleft   = 0;

  // Create picturequeue mutex and condition
  vs->pictq_mutex = SDL_CreateMutex();
  vs->pictq_cond  = SDL_CreateCond();

  // initialize queues
  packet_queue_init(&vs->videoq);
  packet_queue_init(&vs->audioq);

  vs->continue_read_thread = SDL_CreateCond();

  init_clock(&vs->vidclk, &vs->videoq.serial);
  init_clock(&vs->audclk, &vs->audioq.serial);
  init_clock(&vs->extclk, &vs->extclk.serial);
  vs->audio_clock_serial = -1;
  vs->audio_last_serial = -1;

  // Create and run the thread
  vs->decode_tid = SDL_CreateThread(decode_thread, vs);
  if(!vs->decode_tid) {
     av_free(vs);
     return -1;
  }
   imgtime = ICON_0;
   event_loop(vs,sdl_bpp);

  // never return
  return 0;
}

#endif /* HAVE_SDL */
#endif /* ENABLE_PLAYER */

