/********************************************************************************
*                                                                               *
*                      F X   M O V I E   M A N A G E R                          *
*                                                                               *
*********************************************************************************
* Copyright(C) 2008-2014 by Luigi Ziviani <ziviani_luigi@yahoo.it>.             *
* All Rights Reserved.                                 v 6.6                    *
*********************************************************************************
* Id: fxplayer_queue.c,                                2008/01/01               *
*********************************************************************************
* This file is part of FXMOVIEMANAGER                                           *
*                                                                               *
* FXMOVIEMANAGER is free software; you can redistribute it and/or modify        *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation; either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* FXMOVIEMANAGER is distributed in the hope that it will be useful,             *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with this program; if not, write to the Free Software                   *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA     * 
********************************************************************************/

           /* ***************************************************
            * Acknowledgements:                                 *
            * the code written in this file is based on Various * 
            * utilities for command line tools & part of FFmpeg *
            * Copyright (c) 2000-2003 Fabrice Bellard,          *
            *************************************************** */

#include "fxplayer_queue.h"

#ifdef ENABLE_PLAYERS


// packet queue store packet
int packet_queue_put_private(PacketQueue *q, AVPacket *pkt) {

  LocalAVPacketList *pkt_lst;

  if (q->quit)
    return -1;
  // allocate memory
  pkt_lst = av_malloc(sizeof(LocalAVPacketList));
  if (!pkt_lst)
    return -1;

  // populate
  pkt_lst->pkt = *pkt;
  pkt_lst->next = NULL;

  if (pkt == &flush_pkt)
    q->serial++;
  pkt_lst->serial = q->serial;

  if (!q->last_pkt)
    q->first_pkt = pkt_lst;
  else
    q->last_pkt->next = pkt_lst;

  // increment packet queue size
  q->last_pkt = pkt_lst;
  q->nb_packets++;
  q->size += pkt_lst->pkt.size + sizeof(*pkt_lst);
  // should duplicate packet data in DV case
  SDL_CondSignal(q->cond);

  return 0;
}

// packet queue duplicate
int packet_queue_put(PacketQueue *q, AVPacket *pkt) {
  int ret;
  // duplicate packet if not already done 
  if(pkt != &flush_pkt && av_dup_packet(pkt) < 0)
     return -1;

  SDL_LockMutex(q->mutex);
  ret = packet_queue_put_private(q, pkt);
  SDL_UnlockMutex(q->mutex);

  if (pkt != &flush_pkt && ret < 0)
     av_free_packet(pkt);

  return ret;
}

// packet queue initialize
void packet_queue_init(PacketQueue *q) {
  // set the end of buffer to 0 to avoid overlapping
  memset(q, 0, sizeof(PacketQueue));
  q->mutex = SDL_CreateMutex();
  q->cond = SDL_CreateCond();
  q->quit = 1;
}

// packet queue flush
void packet_queue_flush(PacketQueue *q) {

  LocalAVPacketList *pkt_lst, *pkt_lst1;
  // lock
  SDL_LockMutex(q->mutex);
  // free one by one all packets till end
  for(pkt_lst = q->first_pkt; pkt_lst != NULL; pkt_lst = pkt_lst1) {
    pkt_lst1 = pkt_lst->next;
    av_free_packet(&pkt_lst->pkt);
    av_freep(&pkt_lst);
  }
  // reset to initial value
  q->last_pkt = NULL;
  q->first_pkt = NULL;
  q->nb_packets = 0;
  q->size = 0;
  // unlock
  SDL_UnlockMutex(q->mutex);
}

// packet queue destroy
void packet_queue_destroy(PacketQueue *q) {
    packet_queue_flush(q);
    SDL_DestroyMutex(q->mutex);
    SDL_DestroyCond(q->cond);
}

// packet queue lock
void packet_queue_quit(PacketQueue *q) {
    SDL_LockMutex(q->mutex);

    q->quit = 1;
    SDL_CondSignal(q->cond);

    SDL_UnlockMutex(q->mutex);
}

// packet queue start
void packet_queue_start(PacketQueue *q) {
    SDL_LockMutex(q->mutex);

    q->quit = 0;
    packet_queue_put_private(q, &flush_pkt);

    SDL_UnlockMutex(q->mutex);
}

// return -1 if aborted, 0 if no packet and 1 if packet.
int packet_queue_get(PacketQueue *q, AVPacket *pkt, int block, int *serial) {

  LocalAVPacketList *pkt_lst;
  int ret;

  SDL_LockMutex(q->mutex);
  
  for(;;) {
    if(q->quit) {
       ret = -1;
       break;
    }

    pkt_lst = q->first_pkt;
    if (pkt_lst) {
       q->first_pkt = pkt_lst->next;
       if (!q->first_pkt)
          q->last_pkt = NULL;
       q->nb_packets--;
       q->size -= pkt_lst->pkt.size + sizeof(*pkt_lst);
       *pkt = pkt_lst->pkt;
       if (serial)
          *serial = pkt_lst->serial;
       av_free(pkt_lst);
       ret = 1;
       break; 
    } else if (!block) {
       ret = 0;
       break;
    } else {
       SDL_CondWait(q->cond, q->mutex);
    }
  }

  SDL_UnlockMutex(q->mutex);

  return ret;
}



#endif /* ENABLE_PLAYERS */
